import logging
import os
import random
import select
import time
from threading import RLock
from typing import Any, Callable, Literal, Mapping, NamedTuple, Sequence, TypedDict

try:
    # Try to maintain compat with python 3.10
    from typing import NotRequired  # type: ignore
except ImportError:
    from typing import Optional as NotRequired

from .const import Axis, Button, Configuration

logger = logging.getLogger(__name__)

DEBUG_MODE = bool(os.environ.get("HHD_DEBUG", False))


class SpecialEvent(TypedDict):
    type: Literal["special"]
    event: Literal[
        "guide",
        # Builtin Controller
        "qam_single",
        "qam_predouble",
        "qam_double",
        "qam_triple",
        "qam_hold",
        "overlay",
        # Shortcuts
        "xbox_b",
        "xbox_y",
        "kbd_meta_press",
        "kbd_meta_hold",
        "swipe_left_top",
        "swipe_left_bottom",
        "swipe_right_top",
        "swipe_right_bottom",
        "swipe_bottom",
        # TDP Cycle animation
        "tdp_cycle_quiet",
        "tdp_cycle_balanced",
        "tdp_cycle_performance",
        "tdp_cycle_custom",
        # Sleep information
        "wakeup",
        # Powerbutton presses
        "pbtn_short",
        "pbtn_long",
        "pbtn_double",  # todo
        # Debug
        "restart_dev",
        "shutdown_dev",
        "refresh",
    ]
    data: "NotRequired[Any]"


class RumbleEvent(TypedDict):
    """In case ev effects is too complicated. If both magnitudes are 0, disable rumble."""

    type: Literal["rumble"]
    code: Literal["main", "left", "right"]
    strong_magnitude: float
    weak_magnitude: float


RgbMode = Literal["disabled", "solid", "pulse", "rainbow", "spiral", "duality", "oxp"]
RgbSettings = Literal[
    "color", "brightness", "speed", "brightnessd", "speedd", "direction", "dual", "oxp"
]

# Mono is a single zone (main only)
# Dual has per side RGB
# Quad has two zones per stick (Ally)
# TODO: This code needs to be refactored
RgbZones = Literal["mono", "dual", "quad"]
OxpModes = Literal[
    "monster_woke",
    "flowing",
    "sunset",
    "neon",
    "dreamy",
    "cyberpunk",
    "colorful",
    "aurora",
    "sun",
    "classic",
]


class RgbLedEvent(TypedDict):
    """Inspired by new controllers with RGB leds, especially below the buttons.

    Instead of code, this event type exposes multiple properties, including mode."""

    type: Literal["led"]
    initialize: bool

    # Controls the LED zone. Main sets all zones.
    # Left all left zones, right all right zones.
    # One and Two are used for quad zone control and three is used for per side 3 zones.
    # The third zone would be on the bumpers, such as on the Ayn Loki, if it
    # supported per zone RGB.
    code: Literal[
        "main",
        "left",
        "right",
        "left_left",
        "left_right",
        "right_left",
        "right_right",
    ]

    # Various lighting modes supported by the led.
    mode: RgbMode

    # Brightness range is from 0 to 1
    # If the response device does not support brightness control, it shall
    # devide the rgb values by the brightness and round.
    brightness: float

    # The speed the led should blink if supported by the led
    speed: float

    # For the Ally, has three brightness levels
    # (and a forth off, use disabled mode for that)
    brightnessd: Literal["low", "medium", "high"]
    speedd: Literal["low", "medium", "high"]
    direction: Literal["left", "right"]

    # Color values for the led, may be ommited depending on the mode, by being
    # set to 0
    red: int
    green: int
    blue: int

    red2: int
    green2: int
    blue2: int

    oxp: None | OxpModes


class ButtonEvent(TypedDict):
    type: Literal["button"]
    code: Button
    value: bool


class AxisEvent(TypedDict):
    type: Literal["axis"]
    code: Axis
    value: float


class ConfigurationEvent(TypedDict):
    type: Literal["configuration"]
    code: Configuration
    value: Any


class RgbCapabilities(TypedDict):
    modes: dict[RgbMode, Sequence[RgbSettings]] | None
    controller: bool
    zones: RgbZones


class ControllerCapabilities(TypedDict):
    buttons: dict  # TODO
    supports_qam: bool
    rgb: RgbCapabilities | None


Event = ButtonEvent | AxisEvent | ConfigurationEvent | RgbLedEvent | RumbleEvent

GRAB_TIMEOUT = 5

QueueEvent = tuple[Any, Sequence[Event]]


class ControllerEmitter:

    def __init__(self, ctx=None) -> None:
        self.intercept_lock = RLock()
        self._intercept = None
        self._controller_cb = None
        self._qam_cb = None
        self.ctx = ctx
        self._simple_qam = False
        self._cap = None
        self.cid = ""
        self._evs = []

    def send_qam(self, expanded: bool = False):
        with self.intercept_lock:
            if self._qam_cb:
                return self._qam_cb(expanded)
            return False

    def open_steam(self, expanded: bool = False):
        if not self.send_qam(expanded):
            return self.inject(
                {"type": "configuration", "code": "steam", "value": expanded}
            )
        return True

    def set_simple_qam(self, val: bool):
        with self.intercept_lock:
            self._simple_qam = val

    def simple_qam(self):
        with self.intercept_lock:
            return self._simple_qam

    def register_qam(self, cb: Callable[..., bool]):
        with self.intercept_lock:
            self._qam_cb = cb

    def grab(self, enable: bool):
        with self.intercept_lock:
            if enable:
                self._intercept = time.perf_counter()
            else:
                self._intercept = None

    def register_intercept(self, cb: Callable[[Any, Sequence[Event]], None]):
        with self.intercept_lock:
            self._controller_cb = cb

    def should_intercept(self):
        with self.intercept_lock:
            return self._intercept is not None

    def intercept(self, cid: Any, evs: Sequence[Event]):
        with self.intercept_lock:
            if self._intercept:
                if self._intercept + GRAB_TIMEOUT < time.perf_counter():
                    logger.error(
                        f"Intercept timeout triggered, deactivating controller grab."
                    )
                    self.grab(False)
                    return False
                elif evs and self._controller_cb:
                    self._controller_cb(cid, evs)
                    return True
                else:
                    return False
            else:
                return False

    def inject(self, ev: Sequence[Event] | Event):
        if not isinstance(ev, Sequence):
            ev = [ev]
        with self.intercept_lock:
            if not self.cid or (self._cap and not self._cap.get("supports_qam", True)):
                # Avoid writing events if no controller is connected
                return False
            for e in ev:
                self._evs.append((e, 0))
        return True

    def inject_timed(self, evs: Sequence[tuple[Event, float]]):
        # Unfortunately here we have to clear the previous events to avoid conflicts
        # TODO: Clean this up. It is only used by the RGB module.
        with self.intercept_lock:
            self._evs = list(evs)

    def inject_recv(self):
        with self.intercept_lock:
            if not self.cid:
                # Avoid writing events if no controller is connected
                return []

            if not self._evs:
                return []

            curr = time.time()
            removed = []
            tmp = []
            for i, (ev, t) in enumerate(self._evs):
                if curr >= t:
                    tmp.append(ev)
                    removed.insert(0, i)  # prepend to remove in opposite order

            for i in removed:
                self._evs.pop(i)
            return tmp

    def set_capabilities(self, cid, cap: ControllerCapabilities | None):
        with self.intercept_lock:
            self._cap = cap
            self.cid = cid

    def get_capabilities(self) -> dict[str, ControllerCapabilities]:
        with self.intercept_lock:
            if self._cap:
                return {self.cid: self._cap}
            return {}

    def __call__(self, event: SpecialEvent | Sequence[SpecialEvent]) -> None:
        pass


class TouchpadCorrection(NamedTuple):
    x_mult: float = 1
    x_ofs: float = 0
    x_clamp: tuple[float, float] = (0, 1)
    y_mult: float = 1
    y_ofs: float = 0
    y_clamp: tuple[float, float] = (0, 1)


TouchpadCorrectionType = Literal[
    "stretch",
    "crop_center",
    "crop_start",
    "crop_end",
    "contain_start",
    "contain_end",
    "contain_center",
    "left",
    "right",
    "center",
    "disabled",
]


def correct_touchpad(
    width: int, height: int, aspect: float, method: TouchpadCorrectionType
):
    dst = width / height
    src = aspect
    ratio = dst / src

    match method:
        case "left":
            if ratio > 2:
                new_width = width / ratio
                return TouchpadCorrection(
                    x_mult=new_width,
                    x_ofs=0,
                    y_mult=height,
                    y_ofs=0,
                )
            else:
                new_height = height * ratio / 2
                return TouchpadCorrection(
                    x_mult=width / 2,
                    x_ofs=0,
                    y_mult=new_height,
                    y_ofs=(height - new_height),
                )
        case "right":
            if ratio > 2:
                new_width = width / ratio
                return TouchpadCorrection(
                    x_mult=new_width,
                    x_ofs=(width - new_width),
                    y_mult=height,
                    y_ofs=0,
                )
            else:
                new_height = height * ratio / 2
                return TouchpadCorrection(
                    x_mult=width / 2,
                    x_ofs=width / 2,
                    y_mult=new_height,
                    y_ofs=(height - new_height),
                )
        case "center":
            if ratio > 1:
                new_width = width / ratio
                return TouchpadCorrection(
                    x_mult=new_width,
                    x_ofs=(width - new_width) / 2,
                    y_mult=height,
                    y_ofs=0,
                )
            else:
                new_height = height * ratio
                return TouchpadCorrection(
                    x_mult=width,
                    x_ofs=0,
                    y_mult=new_height,
                    y_ofs=(height - new_height) / 2,
                )
        case "crop_center":
            if ratio > 1:
                new_width = width / ratio
                return TouchpadCorrection(
                    x_mult=new_width,
                    x_ofs=(width - new_width) / 2,
                    y_mult=height,
                    y_ofs=0,
                )
            else:
                new_height = height * ratio
                return TouchpadCorrection(
                    x_mult=width,
                    x_ofs=0,
                    y_mult=new_height,
                    y_ofs=(height - new_height) / 2,
                )
        case "crop_start":
            if ratio > 1:
                new_width = width / ratio
                return TouchpadCorrection(
                    x_mult=new_width,
                    x_ofs=0,
                    y_mult=height,
                    y_ofs=0,
                )
            else:
                new_height = height * ratio
                return TouchpadCorrection(
                    x_mult=width,
                    x_ofs=0,
                    y_mult=new_height,
                    y_ofs=0,
                )
        case "crop_end":
            if ratio > 1:
                new_width = width / ratio
                return TouchpadCorrection(
                    x_mult=new_width,
                    x_ofs=(width - new_width),
                    y_mult=height,
                    y_ofs=0,
                )
            else:
                new_height = height * ratio
                return TouchpadCorrection(
                    x_mult=width,
                    x_ofs=0,
                    y_mult=new_height,
                    y_ofs=(height - new_height),
                )
        case "contain_center":
            if ratio > 1:
                bound = (ratio - 1) / ratio / 2
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, y_clamp=(bound, 1 - bound)
                )
            else:
                bound = (1 - ratio) / 2
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, x_clamp=(bound, 1 - bound)
                )
        case "contain_start":
            if ratio > 1:
                bound = (ratio - 1) / ratio
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, y_clamp=(0, 1 - bound)
                )
            else:
                bound = (1 - ratio) / 2
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, x_clamp=(0, 1 - bound)
                )
        case "contain_end":
            if ratio > 1:
                bound = (ratio - 1) / ratio
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, y_clamp=(bound, 1)
                )
            else:
                bound = (1 - ratio) / 2
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, x_clamp=(bound, 1)
                )
        case "stretch" | "disabled":
            return TouchpadCorrection(x_mult=width, y_mult=height)

    logger.error(f"Touchpad correction method '{method}' not found.")
    return TouchpadCorrection(x_mult=width, y_mult=height)


class Producer:
    def open(self) -> Sequence[int]:
        """Opens and returns a list of file descriptors that should be listened to."""
        raise NotImplementedError()

    def close(self, exit: bool) -> bool:
        """Called to close the device.

        If `exit` is true, the program is about to
        close. If it is false, the controller may be performing a configuration
        change.

        In the first versions of Handheld Daemon, this API was meant to be used
        for the controller to enter power saving mode. However, it turns out
        that steam and the kernel do not let the controller disconnect,
        so it was repurposed to skip controller hiding."""
        return False

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        """Called with the file descriptors that are ready to read."""
        return []


class Consumer:
    available: bool
    """Hint that states if the consumer can receive events. If it is false,
    consumer will not be called. If all consumers are false, producers will
    be closed to save CPU utilisation."""

    def initialize(self):
        """Optional method for initialization."""
        pass

    def consume(self, events: Sequence[Event]):
        pass


TouchpadAction = Literal["disabled", "left_click", "right_click"]


class Multiplexer:
    QAM_HOLD_TIME = 0.4
    QAM_MULTI_PRESS_DELAY = 0.2
    QAM_TAP_TIME = 0.04
    QAM_DELAY = 0.15
    REBOOT_HOLD_SELECT = 9
    REBOOT_HOLD_TURBO = 4
    REBOOT_VIBRATION_STRENGTH = 1
    REBOOT_VIBRATION_ON = 0.4
    REBOOT_VIBRATION_OFF = 1.2
    REBOOT_VIBRATION_NUM = 3
    STEAM_CHECK_INTERVAL = 3
    STARTSELECT_TRIGGER_THRESHOLD = 0.6

    def __init__(
        self,
        swap_guide: (
            None
            | Literal[
                "guide_is_start",
                "guide_is_select",
                "select_is_guide",
                "start_is_keyboard",
            ]
        ) = None,
        trigger: None | Literal["analog_to_discrete", "discrete_to_analogue"] = None,
        dpad: (
            None
            | Literal["analog_to_discrete"]
            | Literal["discrete_to_analog"]
            | Literal["both"]
        ) = None,
        led: None | Literal["left_to_main", "right_to_main", "main_to_sides"] = None,
        touchpad: (
            None | Literal["left_to_main", "right_to_main", "main_to_sides"]
        ) = None,
        status: None | Literal["both_to_main"] = None,
        share_to_qam: bool = False,
        trigger_discrete_lvl: float = 0.99,
        touchpad_short: TouchpadAction = "disabled",
        touchpad_right: TouchpadAction = "left_click",
        touchpad_hold: TouchpadAction = "disabled",
        r3_to_share: bool = False,
        select_reboots: bool = False,
        share_reboots: bool = False,
        nintendo_mode: bool = False,
        qam_button: str | None = None,
        emit: ControllerEmitter | None = None,
        imu: None | Literal["left_to_main", "right_to_main", "main_to_sides"] = None,
        params: Mapping[str, Any] = {},
        qam_multi_tap: bool = True,
        qam_no_release: bool = False,
        qam_hhd: bool = False,
        qam_hold: Literal["hhd", "mode"] = "hhd",
        keyboard_is: Literal["steam_qam", "qam", "keyboard"] = "keyboard",
        keyboard_no_release: bool = False,
        startselect_chord: str = "disabled",
    ) -> None:
        self.swap_guide = swap_guide
        self.trigger = trigger
        self.dpad = dpad
        self.led = led
        self.touchpad = touchpad
        self.status = status
        self.trigger_discrete_lvl = trigger_discrete_lvl
        self.touchpad_short = touchpad_short
        self.touchpad_hold = touchpad_hold
        self.touchpad_right = touchpad_right
        self.reboot_button = None
        if select_reboots:
            self.reboot_button = "select"
            self.reboot_time = self.REBOOT_HOLD_SELECT
        if share_reboots:
            self.reboot_button = "share"
            self.reboot_time = self.REBOOT_HOLD_TURBO
        self.r3_to_share = r3_to_share
        self.nintendo_mode = nintendo_mode
        self.emit = emit
        self.send_xbox_b = None
        self.imu = imu
        self.qam_hhd = qam_hhd
        self.qam_hold = qam_hold
        self.keyboard_is = keyboard_is
        self.keyboard_no_release = keyboard_no_release
        self.startselect_chord = startselect_chord
        self.startselect_pressed = None

        self.state = {}
        self.touchpad_x = 0
        self.touchpad_y = 0
        self.touchpad_down = None
        self.queue: list[tuple[Event | Literal["reboot"], float]] = []
        self.reboot_pressed = None
        self.select_is_held = False
        self.reboot_is_held = False
        self.qam_kbd = False
        self.qam_button = qam_button
        if share_to_qam:
            self.qam_button = "share"
        self.has_qam = params.get("has_qam", False)

        self.noob_mode = params.get("noob_mode", False)
        self.qam_pressed = None
        self.qam_pre_sent = False
        self.qam_released = None
        self.qam_times = 0
        self.qam_multi_tap = qam_multi_tap
        self.qam_no_release = qam_no_release
        self.qam_simple = os.environ.get("HHD_QAM_MULTI_DISABLE", None) or (
            self.emit and self.emit.simple_qam()
        )
        self.guide_pressed = False
        self.steam_check = params.get("steam_check", None)
        self.steam_check_last = time.perf_counter()
        self.steam_check_fn = params.get("steam_check_fn", None)
        self.nintendo_qam = params.get("nintendo_qam", False)
        self.open_steam_kbd = params.get("steam_kbd", lambda open: False)

        self.unique = str(time.perf_counter_ns())
        assert touchpad is None, "touchpad rewiring not supported yet"

        uses_rgb: bool = params.get("rgb_used", False)
        rgb_modes: dict[RgbMode, Sequence[RgbSettings]] | None = params.get(
            "rgb_modes", None
        )
        rgb_zones: RgbZones = params.get("rgb_zones", "mono")
        if self.emit:
            rgb = None
            if rgb_modes:
                rgb: RgbCapabilities | None = {
                    "modes": rgb_modes,
                    "controller": uses_rgb,
                    "zones": rgb_zones,
                }
            self.emit.set_capabilities(
                self.unique,
                {
                    "buttons": {},
                    "rgb": rgb,
                    "supports_qam": params.get("supports_qam", True),
                },
            )

    def process(self, events: Sequence[Event]) -> Sequence[Event]:
        out: list[Event] = []
        status_events = set()
        touched = False
        send_steam_qam = False
        send_steam_expand = False

        curr = time.perf_counter()

        # Send old events
        while len(self.queue) and self.queue[0][1] < curr:
            ev = self.queue.pop(0)[0]
            if ev == "reboot":
                if self.reboot_is_held:
                    try:
                        import os

                        os.system("systemctl reboot")
                        logger.info("rebooting")
                    except Exception as e:
                        logger.error(f"Rebooting failed with error:\n{type(e)}:{e}")
            elif self.reboot_is_held or not ev.get("from_reboot", False):
                out.append({**ev, "from_queue": True})  # type: ignore

        # Check for steam for touchpad emulation
        if (
            self.steam_check_fn
            and self.steam_check is not None
            and self.steam_check_last + Multiplexer.STEAM_CHECK_INTERVAL < curr
        ):
            self.steam_check_last = curr
            if self.steam_check:
                msg = "Gamepadui closed. Restarting controller to disable touchpad emulation."
            else:
                msg = "Gamepadui launched. Restarting controller to enable touchpad emulation."

            assert self.steam_check_fn() == self.steam_check, msg

        if self.reboot_pressed and self.reboot_pressed + self.reboot_time < curr:
            self.reboot_pressed = None
            for i in range(self.REBOOT_VIBRATION_NUM):
                self.queue.append(
                    (
                        {
                            "type": "rumble",
                            "code": "main",
                            "strong_magnitude": self.REBOOT_VIBRATION_STRENGTH,
                            "weak_magnitude": self.REBOOT_VIBRATION_STRENGTH,
                            "from_reboot": True,
                        },  # type: ignore
                        curr
                        + i * (self.REBOOT_VIBRATION_ON + self.REBOOT_VIBRATION_OFF),
                    )
                )
                self.queue.append(
                    (
                        {
                            "type": "rumble",
                            "code": "main",
                            "strong_magnitude": 0,
                            "weak_magnitude": 0,
                            "from_reboot": True,
                        },  # type: ignore
                        curr
                        + i * (self.REBOOT_VIBRATION_ON + self.REBOOT_VIBRATION_OFF)
                        + self.REBOOT_VIBRATION_ON,
                    )
                )
            self.queue.append(("reboot", curr))

        if (
            self.touchpad_hold != "disabled"
            and self.touchpad_down
            and self.touchpad_down[3]
            and curr - self.touchpad_down[0] > 0.8
        ):
            action = (
                "touchpad_left"
                if self.touchpad_hold == "left_click"
                else "touchpad_right"
            )
            self.queue.append(
                (
                    {
                        "type": "button",
                        "code": action,
                        "value": True,
                    },
                    curr,
                )
            )
            self.queue.append(
                (
                    {
                        "type": "button",
                        "code": action,
                        "value": False,
                    },
                    curr + self.QAM_DELAY,
                )
            )
            self.touchpad_down = None
        elif self.touchpad_down and (
            abs(self.touchpad_down[1] - self.touchpad_x) > 0.13
            or abs(self.touchpad_down[2] - self.touchpad_y) > 0.13
        ):
            self.touchpad_down[3] = False

        for ev in events:
            match ev["type"]:
                case "axis":
                    match self.imu:
                        case "left_to_main":
                            match ev["code"]:
                                case "left_accel_x":
                                    ev["code"] = "accel_x"
                                case "left_accel_y":
                                    ev["code"] = "accel_y"
                                case "left_accel_z":
                                    ev["code"] = "accel_z"
                                case "left_gyro_x":
                                    ev["code"] = "gyro_x"
                                case "left_gyro_y":
                                    ev["code"] = "gyro_y"
                                case "left_gyro_z":
                                    ev["code"] = "gyro_z"
                                case "left_imu_ts":
                                    ev["code"] = "imu_ts"
                        case "right_to_main":
                            match ev["code"]:
                                case "right_accel_x":
                                    ev["code"] = "accel_x"
                                case "right_accel_y":
                                    ev["code"] = "accel_y"
                                case "right_accel_z":
                                    ev["code"] = "accel_z"
                                case "right_gyro_x":
                                    ev["code"] = "gyro_x"
                                case "right_gyro_y":
                                    ev["code"] = "gyro_y"
                                case "right_gyro_z":
                                    ev["code"] = "gyro_z"
                                case "right_imu_ts":
                                    ev["code"] = "imu_ts"
                        case "main_to_sides":
                            match ev["code"]:
                                case "accel_x":
                                    ev["code"] = "right_accel_x"
                                    ev["code"] = "left_accel_x"
                                case "accel_y":
                                    ev["code"] = "right_accel_y"
                                    ev["code"] = "left_accel_y"
                                case "accel_z":
                                    ev["code"] = "right_accel_z"
                                    ev["code"] = "left_accel_z"
                                case "gyro_x":
                                    ev["code"] = "right_gyro_x"
                                    ev["code"] = "left_gyro_x"
                                case "gyro_y":
                                    ev["code"] = "right_gyro_y"
                                    ev["code"] = "left_gyro_y"
                                case "gyro_z":
                                    ev["code"] = "right_gyro_z"
                                    ev["code"] = "left_gyro_z"
                                case "imu_ts":
                                    ev["code"] = "right_imu_ts"
                                    ev["code"] = "left_imu_ts"

                    if (
                        self.startselect_pressed == "wait"
                        and ev["code"]
                        in (
                            "lt",
                            "rt",
                            "hat_x",
                            "hat_y",
                        )
                        and abs(ev["value"]) > self.STARTSELECT_TRIGGER_THRESHOLD
                    ):
                        out.append(
                            {
                                "type": "button",
                                "code": "mode",
                                "value": True,
                            }
                        )
                        self.startselect_pressed = "pressed"
                    if self.startselect_pressed == "pressed":
                        self.queue.append(
                            (
                                {
                                    "type": "axis",
                                    "code": ev["code"],
                                    "value": ev["value"],
                                },
                                curr + self.QAM_DELAY,
                            )
                        )
                        ev["code"] = ""  # type: ignore

                    if self.trigger == "analog_to_discrete" and ev["code"] in (
                        "lt",
                        "rt",
                    ):
                        out.append(
                            {
                                "type": "button",
                                "code": ev["code"],
                                "value": ev["value"] > self.trigger_discrete_lvl,
                            }
                        )

                    if (
                        self.dpad == "analog_to_discrete" or self.dpad == "both"
                    ) and ev["code"] in (
                        "hat_x",
                        "hat_y",
                    ):
                        out.append(
                            {
                                "type": "button",
                                "code": (
                                    "dpad_down"
                                    if ev["code"] == "hat_y"
                                    else "dpad_right"
                                ),
                                "value": ev["value"] > 0.5,
                            }
                        )
                        out.append(
                            {
                                "type": "button",
                                "code": (
                                    "dpad_up" if ev["code"] == "hat_y" else "dpad_left"
                                ),
                                "value": ev["value"] < -0.5,
                            }
                        )
                    if ev["code"] == "touchpad_x":
                        self.touchpad_x = ev["value"]
                    if ev["code"] == "touchpad_y":
                        self.touchpad_y = ev["value"]
                case "button":
                    if self.trigger == "discrete_to_analog" and ev["code"] in (
                        "lt",
                        "rt",
                    ):
                        out.append(
                            {
                                "type": "axis",
                                "code": ev["code"],
                                "value": 1 if ev["value"] else 0,
                            }
                        )

                    if ev["code"] == "select":
                        if ev["value"]:
                            self.select_is_held = True
                        else:
                            self.select_is_held = False

                    if self.reboot_button and ev["code"] == self.reboot_button:
                        if ev["value"]:
                            self.reboot_pressed = curr
                            self.reboot_is_held = True
                        else:
                            self.reboot_pressed = None
                            self.reboot_is_held = False

                    if self.swap_guide and ev["code"] in (
                        "start",
                        "select",
                        "mode",
                        "share",
                        "keyboard",
                    ):
                        match ev["code"]:
                            # TODO: Refactor the logic of this file,
                            # the arguments do not make sense.
                            case "start":
                                match self.swap_guide:
                                    case "start_is_keyboard":
                                        ev["code"] = "keyboard"
                                    case "select_is_guide":
                                        ev["code"] = "share"
                                    case _:
                                        ev["code"] = "mode"
                            case "select":
                                match self.swap_guide:
                                    case "start_is_keyboard":
                                        ev["code"] = "mode"
                                    case "select_is_guide":
                                        ev["code"] = "mode"
                                    case _:
                                        ev["code"] = "share"
                            case "mode":
                                if self.swap_guide == "guide_is_start":
                                    ev["code"] = "start"
                                else:
                                    ev["code"] = "select"
                            case "share":
                                match self.swap_guide:
                                    case "start_is_keyboard":
                                        pass
                                    case "guide_is_start":
                                        ev["code"] = "select"
                                    case _:
                                        ev["code"] = "start"
                            case "keyboard":
                                if self.swap_guide == "start_is_keyboard":
                                    ev["code"] = "start"

                    if (
                        self.startselect_chord != "disabled" and ev["code"] == "select"
                    ) or (
                        self.startselect_chord == "start_select"
                        and ev["code"] == "start"
                    ):
                        if self.startselect_pressed == "pressed":
                            self.queue.append(
                                (
                                    {
                                        "type": "button",
                                        "code": "mode",
                                        "value": False,
                                    },
                                    curr + self.QAM_DELAY,
                                )
                            )
                            self.startselect_pressed = None

                        if ev["value"]:
                            self.startselect_pressed = "wait"
                        elif self.startselect_pressed == "wait":
                            self.startselect_pressed = None
                            out.append(
                                {
                                    "type": "button",
                                    "code": ev["code"],
                                    "value": True,
                                }
                            )
                            self.queue.append(
                                (
                                    {
                                        "type": "button",
                                        "code": ev["code"],
                                        "value": False,
                                    },
                                    curr + self.QAM_DELAY,
                                )
                            )
                        ev["code"] = ""  # type: ignore

                    if self.emit and ev["code"] == "mode":
                        # Steam might do weirdness, emit an event to prepare
                        # the overlay
                        self.guide_pressed = ev["value"]
                        if ev["value"]:
                            self.emit({"type": "special", "event": "guide"})

                    if (
                        self.dpad == "discrete_to_analog" or self.dpad == "both"
                    ) and ev["code"] in (
                        "dpad_up",
                        "dpad_down",
                        "dpad_left",
                        "dpad_right",
                    ):
                        # FIXME: To be done properly you'd need to save the previous
                        # state so that if going from -1 to 1 in one go it would be
                        # preserved. Since this is only used for the legion go
                        # passthrough that is not an issue.
                        match ev["code"]:
                            case "dpad_up":
                                code = "hat_y"
                                val = -1
                            case "dpad_down":
                                code = "hat_y"
                                val = 1
                            case "dpad_right":
                                code = "hat_x"
                                val = 1
                            case "dpad_left":
                                code = "hat_x"
                                val = -1

                        out.append(
                            {
                                "type": "axis",
                                "code": code,
                                "value": ev["value"] * val,
                            }
                        )

                    if (
                        self.qam_button is not None and ev["code"] == self.qam_button
                    ) or (self.keyboard_is == "qam" and ev["code"] == "keyboard"):
                        self.qam_kbd = ev["code"] == "keyboard"
                        ev["code"] = ""  # type: ignore
                        if not self.qam_simple:
                            if (not self.qam_kbd and self.qam_no_release) or (
                                self.qam_kbd and self.keyboard_no_release
                            ):
                                # Fix for the ally having no hold event
                                if ev["value"]:
                                    self.qam_times += 1
                                    self.qam_released = curr + self.QAM_TAP_TIME
                                    self.qam_pressed = None
                            else:
                                if ev["value"]:
                                    self.qam_times += 1
                                    self.qam_pressed = curr
                                    self.qam_released = None
                                else:
                                    # Only apply if qam_pressed was not yanked
                                    if self.qam_pressed:
                                        self.qam_released = curr
                                    self.qam_pressed = None
                        else:
                            if self.has_qam:
                                out.append(
                                    {
                                        "type": "button",
                                        "code": "share",
                                        "value": ev["value"],
                                    },
                                )
                            else:
                                if ev["value"]:
                                    out.append(
                                        {
                                            "type": "button",
                                            "code": "mode",
                                            "value": True,
                                        },
                                    )
                                    self.queue.append(
                                        (
                                            {
                                                "type": "button",
                                                "code": (
                                                    "b" if self.nintendo_qam else "a"
                                                ),
                                                "value": True,
                                            },
                                            curr + self.QAM_DELAY,
                                        ),
                                    )
                                    self.queue.append(
                                        (
                                            {
                                                "type": "button",
                                                "code": (
                                                    "b" if self.nintendo_qam else "a"
                                                ),
                                                "value": False,
                                            },
                                            curr + 2 * self.QAM_DELAY,
                                        ),
                                    )
                                    self.queue.append(
                                        (
                                            {
                                                "type": "button",
                                                "code": "mode",
                                                "value": False,
                                            },
                                            curr + 2 * self.QAM_DELAY,
                                        ),
                                    )

                    if ev["code"] == "keyboard":
                        if ev["value"]:
                            if self.keyboard_is == "steam_qam":
                                logger.info(f"Keyboard button opens QAM.")
                                send_steam_qam = True
                            elif self.keyboard_is == "keyboard":
                                self.open_steam_kbd(True)
                        ev["code"] = ""

                    if self.noob_mode and ev["code"] == "extra_l1" and ev["value"]:
                        ev["code"] = ""  # type: ignore
                        if self.open_steam_kbd(True):
                            logger.info(f"Opened steam keyboard directly.")
                        else:
                            logger.warning(
                                f"Could not open steam keyboard directly. Sending chord."
                            )
                            out.append(
                                {
                                    "type": "button",
                                    "code": "mode",
                                    "value": True,
                                },
                            )
                            self.queue.append(
                                (
                                    {
                                        "type": "button",
                                        "code": "y" if self.nintendo_qam else "x",
                                        "value": True,
                                    },
                                    curr + self.QAM_DELAY,
                                )
                            )
                            self.queue.append(
                                (
                                    {
                                        "type": "button",
                                        "code": "y" if self.nintendo_qam else "x",
                                        "value": False,
                                    },
                                    curr + 2 * self.QAM_DELAY,
                                ),
                            )
                            self.queue.append(
                                (
                                    {
                                        "type": "button",
                                        "code": "mode",
                                        "value": False,
                                    },
                                    curr + 2 * self.QAM_DELAY,
                                ),
                            )

                    if self.noob_mode and ev["code"] == "extra_r1" and ev["value"]:
                        ev["code"] = ""
                        if self.emit:
                            self.emit({"type": "special", "event": "overlay"})

                    if ev["code"] == "touchpad_right":
                        match self.touchpad_right:
                            case "disabled":
                                # TODO: Cleanup
                                ev["code"] = ""  # type: ignore
                            case "left_click":
                                ev["code"] = "touchpad_left"
                            case "right_click":
                                pass

                    if ev["code"] == "touchpad_touch":
                        if (
                            self.touchpad_short != "disabled"
                            and not ev["value"]
                            and self.touchpad_down
                            and curr - self.touchpad_down[0] < 0.2
                            and abs(self.touchpad_down[1] - self.touchpad_x) < 0.04
                            and abs(self.touchpad_down[2] - self.touchpad_y) < 0.04
                        ):
                            action = (
                                "touchpad_left"
                                if self.touchpad_short == "left_click"
                                else "touchpad_right"
                            )
                            self.queue.append(
                                (
                                    {
                                        "type": "button",
                                        "code": action,
                                        "value": True,
                                    },
                                    curr,
                                )
                            )
                            self.queue.append(
                                (
                                    {
                                        "type": "button",
                                        "code": action,
                                        "value": False,
                                    },
                                    curr + self.QAM_DELAY,
                                )
                            )

                        if ev["value"]:
                            touched = True
                        else:
                            self.touchpad_down = None
                        # append A after QAM_DELAY s

                    if self.r3_to_share and ev["code"] == "extra_r3":
                        ev["code"] = "share"

                    if self.nintendo_mode:
                        match ev["code"]:
                            case "a":
                                ev["code"] = "b"
                            case "b":
                                ev["code"] = "a"
                            case "x":
                                ev["code"] = "y"
                            case "y":
                                ev["code"] = "x"

                    # Assume we own Xbox + Y if the user is not using the recording feature
                    if (
                        (self.guide_pressed or self.select_is_held)
                        and self.emit
                        and ev["code"] == "y"
                        and ev["value"]
                    ):
                        self.emit({"type": "special", "event": "xbox_y"})

                    # Assume we can only use Xbox + B for short presses
                    if (
                        (self.guide_pressed or self.select_is_held)
                        and self.emit
                        and ev["code"] == "b"
                    ):
                        if ev["value"]:
                            self.send_xbox_b = time.time()
                        else:
                            if (
                                self.send_xbox_b
                                and time.time() - self.send_xbox_b < 0.3
                            ):
                                self.emit({"type": "special", "event": "xbox_b"})
                            self.send_xbox_b = None

                    # Apply start/select qam
                    if self.startselect_pressed == "wait" and ev["code"]:
                        out.append(
                            {
                                "type": "button",
                                "code": "mode",
                                "value": True,
                            }
                        )
                        self.startselect_pressed = "pressed"
                    if self.startselect_pressed == "pressed":
                        self.queue.append(
                            (
                                {
                                    "type": "button",
                                    "code": ev["code"],
                                    "value": ev["value"],
                                },
                                curr + self.QAM_DELAY,
                            )
                        )
                        ev["code"] = ""  # type: ignore
                case "led":
                    if self.led == "left_to_main" and ev["code"] == "left":
                        out.append({**ev, "code": "main"})
                    elif self.led == "right_to_main" and ev["code"] == "right":
                        out.append({**ev, "code": "main"})
                    elif self.led == "main_to_both" and ev["code"] == "main":
                        out.append({**ev, "code": "left"})
                        out.append({**ev, "code": "right"})
                case "configuration":
                    if self.status == "both_to_main":
                        self.state[ev["code"]] = ev["value"]
                        match ev["code"]:
                            case "battery_left" | "battery_right":
                                status_events.add("battery")
                            case "is_attached_left" | "is_attached_right":
                                status_events.add("is_attached")
                            case "is_connected_left" | "is_connected_right":
                                status_events.add("is_connected")

        if touched:
            self.touchpad_down = [
                curr,
                self.touchpad_x,
                self.touchpad_y,
                bool(True),
            ]

        for s in status_events:
            match s:
                case "battery":
                    out.append(
                        {
                            "type": "configuration",
                            "code": "battery",
                            "value": min(
                                self.state.get("battery_left", 100),
                                self.state.get("battery_right", 100),
                            ),
                        }
                    )
                case "is_attached":
                    out.append(
                        {
                            "type": "configuration",
                            "code": "is_attached",
                            "value": (
                                self.state.get("is_attached_left", False)
                                and self.state.get("is_attached_right", False)
                            ),
                        }
                    )
                case "is_connected":
                    out.append(
                        {
                            "type": "configuration",
                            "code": "is_connected",
                            "value": (
                                self.state.get("is_connected_left", False)
                                and self.state.get("is_connected_right", False)
                            ),
                        }
                    )

        # Remove empty events
        for ev in events:
            if ev["type"] != "button" or ev["code"]:
                out.append(ev)

        # Handle QAM button
        # Below is the multitap implementation
        # If it was disabled, the code is a NO-OP
        qam_apply = False
        was_held = True
        # Apply hold
        if self.qam_pressed and curr - self.qam_pressed > self.QAM_HOLD_TIME:
            qam_apply = True
        # Apply double tap
        if self.qam_released and (
            curr - self.qam_released > self.QAM_MULTI_PRESS_DELAY
        ):
            qam_apply = True
        # Apply if double tap disabled
        if not self.qam_multi_tap and self.qam_released:
            qam_apply = True

        qam_hhd = self.qam_hhd and not self.qam_kbd
        if (
            self.qam_pressed
            and self.qam_times == (1 if qam_hhd else 2)
            and not self.qam_pre_sent
            and self.emit
        ):
            # Send event instantly after double press to eat delay
            self.emit({"type": "special", "event": "qam_predouble"})
            self.qam_pre_sent = True

        send_steam_qam = send_steam_qam or (
            qam_apply and not qam_hhd and self.qam_released and self.qam_times == 1
        )
        send_steam_expand = (
            qam_apply and self.qam_pressed and was_held and self.qam_hold == "mode"
        )
        if qam_apply and self.emit:
            if qam_hhd:
                match self.qam_times:
                    case 0:
                        pass
                    case 1:
                        self.emit({"type": "special", "event": "qam_double"})
                    case _:
                        self.emit({"type": "special", "event": "qam_triple"})
            else:
                # FIXME: hiding the event based on qam_hold should not happen
                # instead the handler should not open hhd
                if self.qam_pressed and was_held and self.qam_hold == "hhd":
                    self.emit({"type": "special", "event": "qam_hold"})
                else:
                    match self.qam_times:
                        case 0:
                            pass
                        case 1:
                            self.emit({"type": "special", "event": "qam_single"})
                        case 2:
                            self.emit({"type": "special", "event": "qam_double"})
                        case _:
                            self.emit({"type": "special", "event": "qam_triple"})
        if qam_apply:
            held = " then held" if self.qam_pressed else ""
            logger.info(f"QAM Pressed {self.qam_times}{held}.")
            self.qam_pressed = None
            self.qam_released = None
            self.qam_pre_sent = False
            self.qam_times = 0

        if self.emit:
            evs = self.emit.inject_recv()
            # Handle special case for steam
            for ev in evs:
                if ev["type"] == "configuration" and ev["code"] == "steam":
                    if ev["value"]:
                        send_steam_expand = True
                    else:
                        send_steam_qam = True
            out.extend(evs)

        # Grab all events from controller if grab is on
        # Remove queued events such as qam and xbox to avoid leaking them
        # to the overlay
        if self.emit and self.emit.intercept(
            self.unique, [o for o in out if not o.get("from_queue", False)]
        ):
            # Hiding the gyro and sending fake accel values causes
            # issues with emulators and gyro
            # accel = random.random() * 10
            # fake_accel: Sequence[Event] = [
            #     {"type": "axis", "code": "accel_x", "value": accel},
            #     {"type": "axis", "code": "left_accel_x", "value": accel},
            #     {"type": "axis", "code": "right_accel_x", "value": accel},
            # ]
            # return fake_accel + [
            #     o
            #     for o in out
            #     if o["type"] not in ("button", "axis") or "ts" in o.get("code", "")
            # ]
            return [
                o
                for o in out
                if o["type"] not in ("button", "axis")
                or "ts" in o.get("code", "")
                or "accel" in o.get("code", "")
                or "gyro" in o.get("code", "")
            ]
        elif send_steam_qam:
            # Send steam qam only if not intercepting
            if not self.emit or not self.emit.send_qam():
                if self.has_qam:
                    out.append(
                        {
                            "type": "button",
                            "code": "share",
                            "value": True,
                        },
                    )
                    self.queue.append(
                        (
                            {
                                "type": "button",
                                "code": "share",
                                "value": False,
                            },
                            curr + self.QAM_DELAY,
                        )
                    )
                else:
                    # Have a fallback if gamescope is not working
                    out.append(
                        {
                            "type": "button",
                            "code": "mode",
                            "value": True,
                        },
                    )
                    self.queue.append(
                        (
                            {
                                "type": "button",
                                "code": "b" if self.nintendo_qam else "a",
                                "value": True,
                            },
                            curr + self.QAM_DELAY,
                        )
                    )
                    self.queue.append(
                        (
                            {
                                "type": "button",
                                "code": "b" if self.nintendo_qam else "a",
                                "value": False,
                            },
                            curr + 2 * self.QAM_DELAY,
                        ),
                    )
                    self.queue.append(
                        (
                            {
                                "type": "button",
                                "code": "mode",
                                "value": False,
                            },
                            curr + 2 * self.QAM_DELAY,
                        ),
                    )
        elif send_steam_expand:
            out.append(
                {
                    "type": "button",
                    "code": "mode",
                    "value": True,
                },
            )
            self.queue.append(
                (
                    {
                        "type": "button",
                        "code": "mode",
                        "value": False,
                    },
                    curr + self.QAM_DELAY,
                )
            )
        return out


class KeyboardWrapper(Producer, Consumer):
    def __init__(
        self, parent: Producer, button_map: Sequence[tuple[set[Button], Button]]
    ) -> None:
        self.parent = parent
        self.button_map = button_map

        self.active_in: set[Button] = set()
        self.active_out: set[Button] = set()

    def open(self) -> Sequence[int]:
        self.active_in: set[Button] = set()
        self.active_out: set[Button] = set()
        return self.parent.open()

    def close(self, exit: bool) -> bool:
        return self.parent.close(exit)

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        evs: Sequence[Event] = self.parent.produce(fds)
        # Update in map
        for ev in evs:
            logger.info(f"Internal kbd event: {ev}")
            if ev["type"] == "button":
                if ev["value"]:
                    self.active_in.add(ev["code"])
                elif ev["code"] in self.active_in:
                    self.active_in.remove(ev["code"])

        # Debounce and output
        out: Sequence[Event] = []
        for bset, action in self.button_map:
            is_sub = bset.issubset(self.active_in)
            is_active = action in self.active_out
            if is_sub and not is_active:
                self.active_out.add(action)
                out.append({"type": "button", "code": action, "value": True})
            elif not is_sub and is_active:
                self.active_out.remove(action)
                out.append({"type": "button", "code": action, "value": False})

        return out

    def consume(self, events: Sequence[Event]):
        if isinstance(self.parent, Consumer):
            return self.parent.consume(events)


def can_read(fd: int):
    return select.select([fd], [], [], 0)[0]



import re
from typing import Literal, NamedTuple, Sequence


class BM(NamedTuple):
    loc: int
    flipped: bool = False


NumType = Literal["u32", "i32", "m32", "u16", "i16", "m16", "u8", "i8", "m8"]
"""Numerical type for axis.

Number is bit length. Letter signifies sign.
 - `u`: unsigned
 - 'i': signed
 - 'm': signed with middle point. Essentially, if `d` is bit width, `out = in - (1 << d)`
 """


class AM(NamedTuple):
    loc: int
    type: NumType
    order: Literal["little", "big"] = "little"
    scale: float | None = None
    offset: float = 0
    flipped: bool = False
    bounds: tuple[int, int] | None = None


class CM(NamedTuple):
    loc: int
    type: NumType | Literal["bit"]
    order: Literal["little", "big"] = "little"
    scale: float | None = None
    offset: float = 0
    bounds: tuple[int, int] | None = None
    flipped: bool = False


def decode_axis(buff: bytes, t: AM):
    match t.type:
        case "i32":
            o = int.from_bytes(
                buff[t.loc >> 3 : (t.loc >> 3) + 4], t.order, signed=True
            )
            s = (1 << 31) - 1
        case "u32":
            o = int.from_bytes(
                buff[t.loc >> 3 : (t.loc >> 3) + 4], t.order, signed=False
            )
            s = (1 << 32) - 1
        case "m32":
            o = int.from_bytes(
                buff[t.loc >> 3 : (t.loc >> 3) + 4], t.order, signed=False
            ) - (1 << 31)
            s = (1 << 31) - 1
        case "i16":
            o = int.from_bytes(
                buff[t.loc >> 3 : (t.loc >> 3) + 2], t.order, signed=True
            )
            s = (1 << 15) - 1
        case "u16":
            o = int.from_bytes(
                buff[t.loc >> 3 : (t.loc >> 3) + 2], t.order, signed=False
            )
            s = (1 << 16) - 1
        case "m16":
            o = int.from_bytes(
                buff[t.loc >> 3 : (t.loc >> 3) + 2], t.order, signed=False
            ) - (1 << 15)
            s = (1 << 15) - 1
        case "i8":
            o = int.from_bytes(
                buff[t.loc >> 3 : (t.loc >> 3) + 1], t.order, signed=True
            )
            s = (1 << 7) - 1
        case "u8":
            o = int.from_bytes(
                buff[t.loc >> 3 : (t.loc >> 3) + 1], t.order, signed=False
            )
            s = (1 << 8) - 1
        case "m8":
            o = int.from_bytes(
                buff[t.loc >> 3 : (t.loc >> 3) + 1], t.order, signed=False
            ) - (1 << 7)
            s = (1 << 7)
        case _:
            assert False, f"Invalid formatting {t.type}."

    if t.scale:
        v = t.scale * o + t.offset
    else:
        v = o / s + t.offset

    if t.flipped:
        return -v
    else:
        return v


def encode_axis(buff: bytearray, t: AM, val: float):
    if t.flipped:
        val = -val

    if t.scale:
        new_val = int(t.scale * val + t.offset)
        if t.bounds:
            new_val = min(max(new_val, t.bounds[0]), t.bounds[1])
    else:
        new_val = None

    match t.type:
        case "i32":
            if not new_val:
                new_val = int(((1 << 31) - 1) * val)
            buff[t.loc >> 3 : (t.loc >> 3) + 4] = int.to_bytes(
                new_val, 4, t.order, signed=True
            )
        case "u32":
            if not new_val:
                new_val = int(((1 << 32) - 1) * val)
            buff[t.loc >> 3 : (t.loc >> 3) + 4] = int.to_bytes(
                new_val, 4, t.order, signed=False
            )
        case "m32":
            if not new_val:
                new_val = int(round(((1 << 31) - 1) * val + (1 << 31) - 1))
            buff[t.loc >> 3 : (t.loc >> 3) + 4] = int.to_bytes(
                new_val, 4, t.order, signed=False
            )
        case "i16":
            if not new_val:
                new_val = int(((1 << 15) - 1) * val)
            buff[t.loc >> 3 : (t.loc >> 3) + 2] = int.to_bytes(
                new_val, 2, t.order, signed=True
            )
        case "u16":
            if not new_val:
                new_val = int(((1 << 16) - 1) * val)
            buff[t.loc >> 3 : (t.loc >> 3) + 2] = int.to_bytes(
                new_val, 2, t.order, signed=False
            )
        case "m16":
            if not new_val:
                new_val = int(round(((1 << 15) - 1) * val + (1 << 15) - 1))
            buff[t.loc >> 3 : (t.loc >> 3) + 2] = int.to_bytes(
                new_val, 2, t.order, signed=False
            )
        case "i8":
            if not new_val:
                new_val = int(((1 << 7) - 1) * val)
            buff[t.loc >> 3 : (t.loc >> 3) + 1] = int.to_bytes(
                new_val, 1, t.order, signed=True
            )
        case "u8":
            if not new_val:
                new_val = int(((1 << 8) - 1) * val)
            buff[t.loc >> 3 : (t.loc >> 3) + 1] = int.to_bytes(
                new_val, 1, t.order, signed=False
            )
        case "m8":
            if not new_val:
                new_val = int(round(((1 << 7) - 1) * val + (1 << 7) - 1))
            buff[t.loc >> 3 : (t.loc >> 3) + 1] = int.to_bytes(
                new_val, 1, t.order, signed=False
            )
        case _:
            assert False, f"Invalid formatting {t.type}."


def hexify(d: int | Sequence[int]):
    if isinstance(d, int):
        return f"{d:04x}"
    else:
        return [hexify(v) for v in d]


def pretty_print(dev: dict[str, str | int | bytes]):
    out = ""
    for n, v in dev.items():
        if isinstance(v, int):
            out += f"{n}: {hexify(v)}\n"
        elif isinstance(v, str):
            out += f"{n}: '{v}'\n"
        else:
            out += f"{n}: {v}\n"
    return out


def get_button(rep: bytes, map: BM):
    v = bool(rep[map.loc // 8] & (1 << (7 - (map.loc % 8))))
    if map.flipped:
        return not v
    return v


def set_button(rep: bytearray, map: BM, val: bool):
    if (val and not map.flipped) or (not val and map.flipped):
        rep[map.loc // 8] |= 1 << (7 - (map.loc % 8))
    else:
        rep[map.loc // 8] &= 255 - (1 << (7 - (map.loc % 8)))


def decode_config(rep: bytes, map: CM):
    if map.type == "bit":
        return get_button(rep, BM(map.loc, map.flipped))

    ax = decode_axis(rep, AM(map.loc, map.type, map.order, map.scale, map.offset))
    if map.bounds:
        return min(max(ax, map.bounds[0]), map.bounds[1])
    else:
        return ax


def matches_patterns(val: str | int, pats: Sequence[int | str | re.Pattern]):
    if not pats:
        return True

    for p in pats:
        if isinstance(p, re.Pattern):
            assert isinstance(val, str), f"Attempted regex pattern match on val {p}."
            if p.match(val):
                return True
        else:
            if p == val:
                return True
    return False



import logging
import time
from collections import defaultdict
from typing import Sequence

from hhd.controller import Consumer, Event, Producer
from hhd.controller.lib.uhid import UhidDevice, BUS_USB

from .const import (
    SDCONT_VENDOR,
    SDCONT_PRODUCT,
    SDCONT_VERSION,
    SDCONT_COUNTRY,
    SDCONT_NAME,
    SDCONT_DESCRIPTOR,
)

logger = logging.getLogger(__name__)


def trim(rep: bytes):
    if not rep:
        return rep
    idx = len(rep) - 1
    while idx > 0 and rep[idx] == 0x00:
        idx -= 1
    return rep[: idx + 1]


def pad(rep):
    return bytes(rep) + bytes([0 for _ in range(64 - len(rep))])


class SteamdeckOLEDController(Producer, Consumer):
    def __init__(
        self,
    ) -> None:
        self.available = False
        self.report = None
        self.dev = None
        self.start = 0
        self.last_rep = None

    def open(self) -> Sequence[int]:
        self.available = False
        self.report = bytearray([64] + [0 for _ in range(64)])
        self.dev = UhidDevice(
            vid=SDCONT_VENDOR,
            pid=SDCONT_PRODUCT,
            bus=BUS_USB,
            version=SDCONT_VERSION,
            country=SDCONT_COUNTRY,
            name=SDCONT_NAME,
            report_descriptor=SDCONT_DESCRIPTOR,
        )

        self.state: dict = defaultdict(lambda: 0)
        self.rumble = False
        self.touchpad_touch = False
        self.start = time.perf_counter_ns()
        self.fd = self.dev.open()
        return [self.fd]

    def close(self, exit: bool) -> bool:
        if not exit:
            """This is a consumer, so we would deadlock if it was disabled."""
            return False

        if self.dev:
            self.dev.send_destroy()
            self.dev.close()
            self.dev = None
            self.fd = None

        return True

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        if not self.fd or not self.dev or self.fd not in fds:
            return []

        # Process queued events
        out: Sequence[Event] = []
        assert self.dev
        while ev := self.dev.read_event():
            match ev["type"]:
                case "open":
                    logger.info(f"OPENED")
                case "close":
                    logger.info(f"CLOSED")
                case "get_report":
                    match self.last_rep:
                        case 0xAE:
                            rep = bytes(
                                [
                                    0x00,
                                    0xAE,
                                    0x15,
                                    0x01,
                                    *[0x10 for _ in range(15)],
                                ]
                            )
                        case _:
                            rep = bytes([])
                    self.dev.send_get_report_reply(ev["id"], 0, pad(rep))
                    logger.info(
                        f"GET_REPORT: {ev}\nRESPONSE({self.last_rep:02x}): {rep.hex()}"
                    )
                case "set_report":
                    self.dev.send_set_report_reply(ev["id"], 0)
                    logger.info(
                        f"SET_REPORT({ev['rnum']:02x}:{ev['rtype']:02x}): {trim(ev['data']).hex()}"
                    )
                    self.last_rep = ev["data"][3]
                case "output":
                    logger.info(f"OUTPUT")
                case _:
                    logger.warning(f"UKN_EVENT: {ev}")

        return out

    def consume(self, events: Sequence[Event]):
        pass



from evdev import UInput, _uinput, ecodes  # type: ignore
import ctypes
import os


class UInputMonkey(UInput):
    def __init__(
        self,
        events=None,
        name="py-evdev-uinput",
        vendor=0x1,
        product=0x1,
        version=0x1,
        bustype=0x3,
        devnode="/dev/uinput",
        phys="py-evdev-uinput",
        input_props=None,
        uniq=None,
    ):
        self.fd = -1
        try:
            self._new_init(
                events=events,
                name=name,
                vendor=vendor,
                product=product,
                version=version,
                bustype=bustype,
                devnode=devnode,
                phys=phys,
                input_props=input_props,
                uniq=uniq,
            )
        except Exception as e:
            if self.fd != -1:
                os.close(self.fd)
            raise e

    def _new_init(
        self,
        events=None,
        name="py-evdev-uinput",
        vendor=0x1,
        product=0x1,
        version=0x1,
        bustype=0x3,
        devnode="/dev/uinput",
        phys="py-evdev-uinput",
        input_props=None,
        uniq=None,
    ):
        self.name = name  #: Uinput device name.
        self.vendor = vendor  #: Device vendor identifier.
        self.product = product  #: Device product identifier.
        self.version = version  #: Device version identifier.
        self.bustype = bustype  #: Device bustype - e.g. ``BUS_USB``.
        self.phys = phys  #: Uinput device physical path.
        self.devnode = devnode  #: Uinput device node - e.g. ``/dev/uinput/``.

        if not events:
            events = {ecodes.EV_KEY: ecodes.keys.keys()}  # type: ignore

        self._verify()

        #: Write-only, non-blocking file descriptor to the uinput device node.
        self.fd = _uinput.open(devnode)

        # Prepare the list of events for passing to _uinput.enable and _uinput.setup.
        absinfo, prepared_events = self._prepare_events(events)

        # Set phys name
        _uinput.set_phys(self.fd, phys)

        # Set properties
        input_props = input_props or []
        for prop in input_props:
            _uinput.set_prop(self.fd, prop)

        for etype, code in prepared_events:
            _uinput.enable(self.fd, etype, code)

        try:
            _uinput.setup(self.fd, name, vendor, product, version, bustype, absinfo, ecodes.FF_MAX_EFFECTS)  # type: ignore
        except TypeError:
            _uinput.setup(self.fd, name, vendor, product, version, bustype, absinfo)

        if uniq:
            from fcntl import ioctl
            from ...lib.ioctl import UI_SET_UNIQ_STR

            c_uniq = ctypes.create_string_buffer(uniq.encode())
            ioctl(self.fd, UI_SET_UNIQ_STR(len(c_uniq)), c_uniq, False)

        # Create the uinput device.
        _uinput.create(self.fd)

        self.dll = ctypes.CDLL(_uinput.__file__)
        self.dll._uinput_begin_upload.restype = ctypes.c_int
        self.dll._uinput_end_upload.restype = ctypes.c_int

        self.device = None

    def _find_device(self, fd):
        return None



import logging
import time
from typing import Literal, Sequence, cast

from hhd.controller import DEBUG_MODE, Event, RgbMode
from hhd.plugins import Config, Context, HHDPlugin, load_relative_yaml
from hhd.utils import get_distro_color, hsb_to_rgb

logger = logging.getLogger(__name__)

RGB_SET_TIMES = 2
RGB_SET_INTERVAL = 5
RGB_MIN_INTERVAL = 0.1
RGB_QUEUE_RGB = 1.5


class RgbPlugin(HHDPlugin):
    def __init__(self) -> None:
        self.name = f"controller_rgb"
        self.priority = 15
        self.log = "LEDS"

        self.modes = None
        self.controller = False
        self.loaded = False
        self.enabled = False
        self.last_set = 0
        self.queue_leds = None

        self.init_count = 0
        self.init_last = 0
        self.init = True
        self.uniq = None
        self.restore = None
        self.last_ev = None

        self.prev = None

    def open(
        self,
        emit,
        context: Context,
    ):
        self.started = False
        self.emit = emit

    def notify(self, events):
        for ev in events:
            # Certain ayaneo devices reset LEDs when being
            # plugged in
            if ev["type"] == "acpi" and ev["event"] in ("ac", "dc"):
                self.init = True
                self.init_count = RGB_SET_TIMES - 1
            elif ev["type"] == "special":
                match ev["event"]:
                    case "tdp_cycle_quiet":
                        color = (0, 0, 255)
                    case "tdp_cycle_balanced":
                        color = (255, 255, 255)
                    case "tdp_cycle_performance":
                        color = (255, 0, 0)
                    case "tdp_cycle_custom":
                        color = (157, 0, 255)
                    case _:
                        color = None

                if color:
                    red, green, blue = color
                    curr = time.time()
                    evs: Sequence[tuple[Event, float]] = []
                    # Set color based on mode on low brightness
                    if not self.controller:
                        evs.append(
                            (
                                {
                                    "type": "led",
                                    "initialize": True,  # Always initialize, saves problems on the ally
                                    "code": "main",
                                    "mode": "solid",
                                    "direction": "left",
                                    "brightness": 0.33,
                                    "brightnessd": "low",
                                    "speed": 0,
                                    "speedd": "low",
                                    "red": red,
                                    "green": green,
                                    "blue": blue,
                                    "red2": 0,
                                    "green2": 0,
                                    "blue2": 0,
                                    "oxp": None,
                                },
                                curr,
                            ),
                        )
                    # Add short vibration
                    evs.append(
                        (
                            {
                                "type": "rumble",
                                "code": "main",
                                "strong_magnitude": 0.2,
                                "weak_magnitude": 0.2,
                            },
                            curr,
                        ),
                    )
                    evs.append(
                        (
                            {
                                "type": "rumble",
                                "code": "main",
                                "strong_magnitude": 0,
                                "weak_magnitude": 0,
                            },
                            curr + 0.1,
                        ),
                    )
                    # Restore old color
                    if not self.controller and self.last_ev:
                        evs.append((self.last_ev, curr + 3))

                    self.emit.inject_timed(evs)

    def settings(self):
        if not self.modes:
            self.loaded = False
            return {}
        self.loaded = True

        # If RGB support is disabled
        # return enable option only
        base = load_relative_yaml("settings.yml")
        if not self.enabled:
            del base["rgb"]
            return base

        if self.controller:
            # Remove RGB settings because the controller has control
            del base["rgb"]["handheld"]["children"]["mode"]
        else:
            # Remove disclaimer
            del base["rgb"]["handheld"]["children"]["controller"]
            modes = load_relative_yaml("modes.yml")
            capabilities = load_relative_yaml("capabilities.yml")

            # Set a sane default color
            dc = get_distro_color()

            supported = {}
            for mode, caps in self.modes.items():
                if mode in modes:
                    m = modes[mode]
                    m["children"] = {}
                    for cap in caps:
                        m["children"].update(
                            {k: dict(v) for k, v in capabilities[cap].items()}
                        )
                        if cap == "color":
                            m["children"]["hue"]["default"] = dc
                    for c in m["children"].values():
                        c["tags"] = sorted(set(c.get("tags", []) + m.get("tags", [])))
                    supported[mode] = m

            # Add supported modes
            base["rgb"]["handheld"]["children"]["mode"]["modes"] = supported

            # Set a sane default mode
            for default in ("solid", "pulse", "disabled"):
                if default in supported:
                    base["rgb"]["handheld"]["children"]["mode"]["default"] = default
                    break
            else:
                # fallback to any supported mode to have persistence in the mode
                base["rgb"]["handheld"]["children"]["mode"]["default"] = next(
                    iter(supported)
                )
        return base

    def update(self, conf: Config):
        cap = self.emit.get_capabilities()

        if DEBUG_MODE:
            cap = {
                "_dbg": {
                    "rgb": {
                        "controller": False,
                        "modes": {
                            "disabled": [],
                            "solid": ["color"],
                            "pulse": ["color", "speed", "speedd"],
                            "duality": ["dual", "speedd"],
                            "rainbow": ["brightness", "speed", "speedd"],
                            "spiral": ["brightness", "speed", "speedd", "direction"],
                        },
                    }
                }
            }

        if not cap:
            if self.modes:
                self.modes = None
                self.emit({"type": "settings"})
            return

        # Check controller id and force setting the leds if it changed.
        # This will reset the led color after suspend or after exitting
        # dualsense emulation.
        uniq = next(iter(cap))
        ccap = cap[uniq]
        if uniq != self.uniq:
            self.prev = None
            self.uniq = uniq

        rgb = ccap["rgb"]
        refresh_settings = False
        if rgb:
            # Refresh on initial load
            if not self.modes:
                refresh_settings = True

            # Refresh if controller takes control of the LEDs
            new_controller = rgb["controller"]
            if self.controller != new_controller:
                refresh_settings = True

            self.controller = new_controller
            self.modes = rgb["modes"]

        if self.loaded:
            new_enabled = conf["hhd"]["settings"]["rgb"].to(bool)
            if new_enabled != self.enabled:
                refresh_settings = True
                self.enabled = new_enabled

        if refresh_settings:
            self.init_count = 0
            self.init_last = 0
            self.emit({"type": "settings"})
            return

        # All checks were ran and settings were updated
        # if the controller has control of the LEDs
        # or they are not enabled, exit.
        if not self.enabled or self.controller:
            return

        curr = time.perf_counter()
        init = False

        rgb_conf = conf["rgb"]["handheld"]["mode"]
        if self.prev and self.prev != rgb_conf:
            self.init = False
        elif self.init:
            # Initialize by setting the LEDs X times
            # to avoid early boot having it not set
            if self.init_count >= RGB_SET_TIMES:
                self.init = False
                return

            # Wait inbetween setting the LEDS
            curr = time.perf_counter()
            if curr - self.init_last < RGB_SET_INTERVAL:
                return

            self.init_count += 1
            self.init_last = curr
            logger.info(
                f"Initializing RGB (repeat {self.init_count}/{RGB_SET_TIMES}, interval: {RGB_SET_INTERVAL})"
            )
            init = True
        elif self.queue_leds and self.queue_leds < curr:
            # Set the LEDs after two seconds with init
            logger.info("Running full rgb command.")
            init = True
        elif self.prev and self.prev == rgb_conf:
            return

        self.prev = rgb_conf.copy()

        # Get event info
        mode = rgb_conf["mode"].to(str)
        if mode in rgb_conf:
            info = cast(dict, rgb_conf[mode].conf)
        else:
            info = {}
        ev: Event | None = None
        if not self.modes or mode not in self.modes:
            return

        brightness = 1
        brightnessd = "high"
        speedd = "high"
        direction = "left"
        speed = 1
        red = 0
        green = 0
        blue = 0
        red2 = 0
        green2 = 0
        blue2 = 0
        color2_set = False
        always_init = True
        oxp = None

        log = f"Setting RGB to mode '{mode}'"
        for cap in self.modes[cast(RgbMode, mode)]:
            match cap:
                case "color":
                    red, green, blue = hsb_to_rgb(
                        info["hue"],
                        info["saturation"],
                        info["brightness"],
                    )
                    # Cannot init leds with color slider because it is too fast
                    always_init = False
                    log += f" with color: {red:3d}, {green:3d}, {blue:3d}"
                case "dual":
                    red, green, blue = hsb_to_rgb(
                        info["hue"],
                        info["saturation"],
                        info["brightness"],
                    )
                    color2_set = True
                    red2, green2, blue2 = hsb_to_rgb(
                        info["hue2"],
                        info["saturation"],
                        info["brightness"],
                    )
                    # Cannot init leds with color slider because it is too fast
                    always_init = False
                    log += f" with colors: {red:3d}, {green:3d}, {blue:3d} and {red2:3d}, {green2:3d}, {blue2:3d}"
                case "brightness":
                    log += f", brightness: {info['brightness']}"
                    brightness = info["brightness"] / 100
                case "speed":
                    log += f", speed: {info['speed']}"
                    speed = info["speed"] / 100
                case "brightnessd":
                    log += f", brightness: {info['brightnessd']}"
                    brightnessd = cast(
                        Literal["low", "medium", "high"], info["brightnessd"]
                    )
                case "speedd":
                    log += f", speed: {info['speedd']}"
                    speedd = cast(Literal["low", "medium", "high"], info["speedd"])
                case "direction":
                    log += f", direction: {info['direction']}"
                    direction = cast(Literal["left", "right"], info["direction"])
                case "oxp":
                    brightnessd = cast(
                        Literal["low", "medium", "high"], info["brightnessd"]
                    )
                    log += f", mode: '{info['mode']}', brightness: '{brightnessd}'"
                    oxp = info["mode"]
                case "oxp-secondary":
                    log += f", center hue: {info['hue']}, enabled: {info['secondary']}"
                    if info["secondary"]:
                        red2, green2, blue2 = hsb_to_rgb(
                            info["hue"],
                            100,
                            100,
                        )
                    else:
                        red2 = green2 = blue2 = 0
                    color2_set = True

        log += "."

        if not color2_set:
            red2 = red
            green2 = green
            blue2 = blue

        ev = {
            "type": "led",
            "initialize": init
            or always_init,  # Always initialize, saves problems on the ally
            "code": "main",
            "mode": cast(RgbMode, mode),
            "direction": direction,
            "brightness": brightness,
            "brightnessd": brightnessd,
            "speed": speed,
            "speedd": speedd,
            "red": red,
            "green": green,
            "blue": blue,
            "red2": red2,
            "green2": green2,
            "blue2": blue2,
            "oxp": oxp,
        }
        if not always_init:
            self.queue_leds = curr + RGB_QUEUE_RGB

        # Avoid setting the LEDs too fast.
        if curr - self.last_set < RGB_MIN_INTERVAL and not init:
            return

        logger.info(log)
        self.last_set = curr
        self.last_ev = ev
        self.emit.inject(ev)
        if init:
            self.queue_leds = None


def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if len(existing):
        return existing

    return [RgbPlugin()]



"""
Module for deserializing/serializing to and from VDF

SPDX-License-Identifier: MIT
From https://github.com/ValvePython/vdf/
with https://github.com/ValvePython/vdf/pull/61 from Matoking
"""
__version__ = "3.4"
__author__ = "Rossen Georgiev"

import re
import struct
import sys
from binascii import crc32
from collections.abc import Mapping
from io import BytesIO
from io import StringIO as unicodeIO

from .vdict import VDFDict

# Py2 & Py3 compatibility
if sys.version_info[0] >= 3:
    string_type = str
    int_type = int
    BOMS = '\ufffe\ufeff'

    def strip_bom(line):
        return line.lstrip(BOMS)
else:
    from StringIO import StringIO as strIO
    string_type = basestring
    int_type = long
    BOMS = '\xef\xbb\xbf\xff\xfe\xfe\xff'
    BOMS_UNICODE = '\\ufffe\\ufeff'.decode('unicode-escape')

    def strip_bom(line):
        return line.lstrip(BOMS if isinstance(line, str) else BOMS_UNICODE)

# string escaping
_unescape_char_map = {
    r"\n": "\n",
    r"\t": "\t",
    r"\v": "\v",
    r"\b": "\b",
    r"\r": "\r",
    r"\f": "\f",
    r"\a": "\a",
    r"\\": "\\",
    r"\?": "?",
    r"\"": "\"",
    r"\'": "\'",
}
_escape_char_map = {v: k for k, v in _unescape_char_map.items()}

def _re_escape_match(m):
    return _escape_char_map[m.group()]

def _re_unescape_match(m):
    return _unescape_char_map[m.group()]

def _escape(text):
    return re.sub(r"[\n\t\v\b\r\f\a\\\?\"']", _re_escape_match, text)

def _unescape(text):
    return re.sub(r"(\\n|\\t|\\v|\\b|\\r|\\f|\\a|\\\\|\\\?|\\\"|\\')", _re_unescape_match, text)

# parsing and dumping for KV1
def parse(fp, mapper=dict, merge_duplicate_keys=True, escaped=True):
    """
    Deserialize ``s`` (a ``str`` or ``unicode`` instance containing a VDF)
    to a Python object.

    ``mapper`` specifies the Python object used after deserializetion. ``dict` is
    used by default. Alternatively, ``collections.OrderedDict`` can be used if you
    wish to preserve key order. Or any object that acts like a ``dict``.

    ``merge_duplicate_keys`` when ``True`` will merge multiple KeyValue lists with the
    same key into one instead of overwriting. You can se this to ``False`` if you are
    using ``VDFDict`` and need to preserve the duplicates.
    """
    if not issubclass(mapper, Mapping):
        raise TypeError("Expected mapper to be subclass of dict, got %s" % type(mapper))
    if not hasattr(fp, 'readline'):
        raise TypeError("Expected fp to be a file-like object supporting line iteration")

    stack = [mapper()]
    expect_bracket = False

    re_keyvalue = re.compile(r'^("(?P<qkey>(?:\\.|[^\\"])*)"|(?P<key>#?[a-z0-9\-\_\\\?$%<>]+))'
                             r'([ \t]*('
                             r'"(?P<qval>(?:\\.|[^\\"])*)(?P<vq_end>")?'
                             r'|(?P<val>(?:(?<!/)/(?!/)|[a-z0-9\-\_\\\?\*\.$<> ])+)'
                             r'|(?P<sblock>{[ \t]*)(?P<eblock>})?'
                             r'))?',
                             flags=re.I)

    for lineno, line in enumerate(fp, 1):
        if lineno == 1:
            line = strip_bom(line)

        line = line.lstrip()

        # skip empty and comment lines
        if line == "" or line[0] == '/':
            continue

        # one level deeper
        if line[0] == "{":
            expect_bracket = False
            continue

        if expect_bracket:
            raise SyntaxError("vdf.parse: expected openning bracket",
                              (getattr(fp, 'name', '<%s>' % fp.__class__.__name__), lineno, 1, line))

        # one level back
        if line[0] == "}":
            if len(stack) > 1:
                stack.pop()
                continue

            raise SyntaxError("vdf.parse: one too many closing parenthasis",
                              (getattr(fp, 'name', '<%s>' % fp.__class__.__name__), lineno, 0, line))

        # parse keyvalue pairs
        while True:
            match = re_keyvalue.match(line)

            if not match:
                try:
                    line += next(fp)
                    continue
                except StopIteration:
                    raise SyntaxError("vdf.parse: unexpected EOF (open key quote?)",
                                      (getattr(fp, 'name', '<%s>' % fp.__class__.__name__), lineno, 0, line))

            key = match.group('key') if match.group('qkey') is None else match.group('qkey')
            val = match.group('qval')
            if val is None:
                val = match.group('val')
                if val is not None:
                    val = val.rstrip()
                    if val == "":
                        val = None

            if escaped:
                key = _unescape(key)

            # we have a key with value in parenthesis, so we make a new dict obj (level deeper)
            if val is None:
                if merge_duplicate_keys and key in stack[-1]:
                    _m = stack[-1][key]
                    # we've descended a level deeper, if value is str, we have to overwrite it to mapper
                    if not isinstance(_m, mapper):
                        _m = stack[-1][key] = mapper()
                else:
                    _m = mapper()
                    stack[-1][key] = _m

                if match.group('eblock') is None:
                    # only expect a bracket if it's not already closed or on the same line
                    stack.append(_m)
                    if match.group('sblock') is None:
                        expect_bracket = True

            # we've matched a simple keyvalue pair, map it to the last dict obj in the stack
            else:
                # if the value is line consume one more line and try to match again,
                # until we get the KeyValue pair
                if match.group('vq_end') is None and match.group('qval') is not None:
                    try:
                        line += next(fp)
                        continue
                    except StopIteration:
                        raise SyntaxError("vdf.parse: unexpected EOF (open quote for value?)",
                                          (getattr(fp, 'name', '<%s>' % fp.__class__.__name__), lineno, 0, line))

                stack[-1][key] = _unescape(val) if escaped else val

            # exit the loop
            break

    if len(stack) != 1:
        raise SyntaxError("vdf.parse: unclosed parenthasis or quotes (EOF)",
                           (getattr(fp, 'name', '<%s>' % fp.__class__.__name__), lineno, 0, line))

    return stack.pop()


def loads(s, **kwargs):
    """
    Deserialize ``s`` (a ``str`` or ``unicode`` instance containing a JSON
    document) to a Python object.
    """
    if not isinstance(s, string_type):
        raise TypeError("Expected s to be a str, got %s" % type(s))

    fp = unicodeIO(s)
    return parse(fp, **kwargs)


def load(fp, **kwargs):
    """
    Deserialize ``fp`` (a ``.readline()``-supporting file-like object containing
    a JSON document) to a Python object.
    """
    return parse(fp, **kwargs)


def dumps(obj, pretty=False, escaped=True):
    """
    Serialize ``obj`` to a VDF formatted ``str``.
    """
    if not isinstance(obj, Mapping):
        raise TypeError("Expected data to be an instance of``dict``")
    if not isinstance(pretty, bool):
        raise TypeError("Expected pretty to be of type bool")
    if not isinstance(escaped, bool):
        raise TypeError("Expected escaped to be of type bool")

    return ''.join(_dump_gen(obj, pretty, escaped))


def dump(obj, fp, pretty=False, escaped=True):
    """
    Serialize ``obj`` as a VDF formatted stream to ``fp`` (a
    ``.write()``-supporting file-like object).
    """
    if not isinstance(obj, Mapping):
        raise TypeError("Expected data to be an instance of``dict``")
    if not hasattr(fp, 'write'):
        raise TypeError("Expected fp to have write() method")
    if not isinstance(pretty, bool):
        raise TypeError("Expected pretty to be of type bool")
    if not isinstance(escaped, bool):
        raise TypeError("Expected escaped to be of type bool")

    for chunk in _dump_gen(obj, pretty, escaped):
        fp.write(chunk)


def _dump_gen(data, pretty=False, escaped=True, level=0):
    indent = "\t"
    line_indent = ""

    if pretty:
        line_indent = indent * level

    for key, value in data.items():
        if escaped and isinstance(key, string_type):
            key = _escape(key)

        if isinstance(value, Mapping):
            yield '%s"%s"\n%s{\n' % (line_indent, key, line_indent)
            for chunk in _dump_gen(value, pretty, escaped, level+1):
                yield chunk
            yield "%s}\n" % line_indent
        else:
            if escaped and isinstance(value, string_type):
                value = _escape(value)

            yield '%s"%s" "%s"\n' % (line_indent, key, value)


# binary VDF
class BASE_INT(int_type):
    def __repr__(self):
        return "%s(%d)" % (self.__class__.__name__, self)

class UINT_64(BASE_INT):
    pass

class INT_64(BASE_INT):
    pass

class POINTER(BASE_INT):
    pass

class COLOR(BASE_INT):
    pass

BIN_NONE        = b'\x00'
BIN_STRING      = b'\x01'
BIN_INT32       = b'\x02'
BIN_FLOAT32     = b'\x03'
BIN_POINTER     = b'\x04'
BIN_WIDESTRING  = b'\x05'
BIN_COLOR       = b'\x06'
BIN_UINT64      = b'\x07'
BIN_END         = b'\x08'
BIN_INT64       = b'\x0A'
BIN_END_ALT     = b'\x0B'

def binary_loads(b, mapper=dict, merge_duplicate_keys=True, alt_format=False, key_table=None, raise_on_remaining=True):
    """
    Deserialize ``b`` (``bytes`` containing a VDF in "binary form")
    to a Python object.

    ``mapper`` specifies the Python object used after deserializetion. ``dict` is
    used by default. Alternatively, ``collections.OrderedDict`` can be used if you
    wish to preserve key order. Or any object that acts like a ``dict``.

    ``merge_duplicate_keys`` when ``True`` will merge multiple KeyValue lists with the
    same key into one instead of overwriting. You can se this to ``False`` if you are
    using ``VDFDict`` and need to preserve the duplicates.

    ``key_table`` will be used to translate keys in binary VDF objects
    which do not encode strings directly but instead store them in an out-of-band
    table. Newer `appinfo.vdf` format stores this table the end of the file,
    and it is needed to deserialize the binary VDF objects in that file.
    """
    if not isinstance(b, bytes):
        raise TypeError("Expected s to be bytes, got %s" % type(b))

    return binary_load(BytesIO(b), mapper, merge_duplicate_keys, alt_format, key_table, raise_on_remaining)

def binary_load(fp, mapper=dict, merge_duplicate_keys=True, alt_format=False, key_table=None, raise_on_remaining=False):
    """
    Deserialize ``fp`` (a ``.read()``-supporting file-like object containing
    binary VDF) to a Python object.

    ``mapper`` specifies the Python object used after deserializetion. ``dict` is
    used by default. Alternatively, ``collections.OrderedDict`` can be used if you
    wish to preserve key order. Or any object that acts like a ``dict``.

    ``merge_duplicate_keys`` when ``True`` will merge multiple KeyValue lists with the
    same key into one instead of overwriting. You can se this to ``False`` if you are
    using ``VDFDict`` and need to preserve the duplicates.

    ``key_table`` will be used to translate keys in binary VDF objects
    which do not encode strings directly but instead store them in an out-of-band
    table. Newer `appinfo.vdf` format stores this table the end of the file,
    and it is needed to deserialize the binary VDF objects in that file.
    """
    if not hasattr(fp, 'read') or not hasattr(fp, 'tell') or not hasattr(fp, 'seek'):
        raise TypeError("Expected fp to be a file-like object with tell()/seek() and read() returning bytes")
    if mapper is None:
        mapper = dict
    if not issubclass(mapper, Mapping):
        raise TypeError("Expected mapper to be subclass of dict, got %s" % type(mapper))

    # helpers
    int32 = struct.Struct('<i')
    uint64 = struct.Struct('<Q')
    int64 = struct.Struct('<q')
    float32 = struct.Struct('<f')

    def read_string(fp, wide=False):
        buf, end = b'', -1
        offset = fp.tell()

        # locate string end
        while end == -1:
            chunk = fp.read(64)

            if chunk == b'':
                raise SyntaxError("Unterminated cstring (offset: %d)" % offset)

            buf += chunk
            end = buf.find(b'\x00\x00' if wide else b'\x00')

        if wide:
            end += end % 2

        # rewind fp
        fp.seek(end - len(buf) + (2 if wide else 1), 1)

        # decode string
        result = buf[:end]

        if wide:
            result = result.decode('utf-16')
        elif bytes is not str:
            result = result.decode('utf-8', 'replace')
        else:
            try:
                result.decode('ascii')
            except:
                result = result.decode('utf-8', 'replace')

        return result

    stack = [mapper()]
    CURRENT_BIN_END = BIN_END if not alt_format else BIN_END_ALT

    for t in iter(lambda: fp.read(1), b''):
        if t == CURRENT_BIN_END:
            if len(stack) > 1:
                stack.pop()
                continue
            break

        if key_table:
            # If 'key_table' was provided, each key is an int32 value that
            # needs to be mapped to an actual field name using a key table.
            # Newer appinfo.vdf (V29+) stores this table at the end of the file.
            index = int32.unpack(fp.read(int32.size))[0]

            key = key_table[index]
        else:
            key = read_string(fp)

        if t == BIN_NONE:
            if merge_duplicate_keys and key in stack[-1]:
                _m = stack[-1][key]
            else:
                _m = mapper()
                stack[-1][key] = _m
            stack.append(_m)
        elif t == BIN_STRING:
            stack[-1][key] = read_string(fp)
        elif t == BIN_WIDESTRING:
            stack[-1][key] = read_string(fp, wide=True)
        elif t in (BIN_INT32, BIN_POINTER, BIN_COLOR):
            val = int32.unpack(fp.read(int32.size))[0]

            if t == BIN_POINTER:
                val = POINTER(val)
            elif t == BIN_COLOR:
                val = COLOR(val)

            stack[-1][key] = val
        elif t == BIN_UINT64:
            stack[-1][key] = UINT_64(uint64.unpack(fp.read(int64.size))[0])
        elif t == BIN_INT64:
            stack[-1][key] = INT_64(int64.unpack(fp.read(int64.size))[0])
        elif t == BIN_FLOAT32:
            stack[-1][key] = float32.unpack(fp.read(float32.size))[0]
        else:
            raise SyntaxError("Unknown data type at offset %d: %s" % (fp.tell() - 1, repr(t)))

    if len(stack) != 1:
        raise SyntaxError("Reached EOF, but Binary VDF is incomplete")
    if raise_on_remaining and fp.read(1) != b'':
        fp.seek(-1, 1)
        raise SyntaxError("Binary VDF ended at offset %d, but there is more data remaining" % (fp.tell() - 1))

    return stack.pop()

def binary_dumps(obj, alt_format=False):
    """
    Serialize ``obj`` to a binary VDF formatted ``bytes``.
    """
    buf = BytesIO()
    binary_dump(obj, buf, alt_format)
    return buf.getvalue()

def binary_dump(obj, fp, alt_format=False):
    """
    Serialize ``obj`` to a binary VDF formatted ``bytes`` and write it to ``fp`` filelike object
    """
    if not isinstance(obj, Mapping):
        raise TypeError("Expected obj to be type of Mapping")
    if not hasattr(fp, 'write'):
        raise TypeError("Expected fp to have write() method")

    for chunk in _binary_dump_gen(obj, alt_format=alt_format):
        fp.write(chunk)

def _binary_dump_gen(obj, level=0, alt_format=False):
    if level == 0 and len(obj) == 0:
        return

    int32 = struct.Struct('<i')
    uint64 = struct.Struct('<Q')
    int64 = struct.Struct('<q')
    float32 = struct.Struct('<f')

    for key, value in obj.items():
        if isinstance(key, string_type):
            key = key.encode('utf-8')
        else:
            raise TypeError("dict keys must be of type str, got %s" % type(key))

        if isinstance(value, Mapping):
            yield BIN_NONE + key + BIN_NONE
            for chunk in _binary_dump_gen(value, level+1, alt_format=alt_format):
                yield chunk
        elif isinstance(value, UINT_64):
            yield BIN_UINT64 + key + BIN_NONE + uint64.pack(value)
        elif isinstance(value, INT_64):
            yield BIN_INT64 + key + BIN_NONE + int64.pack(value)
        elif isinstance(value, string_type):
            try:
                value = value.encode('utf-8') + BIN_NONE
                yield BIN_STRING
            except:
                value = value.encode('utf-16') + BIN_NONE*2
                yield BIN_WIDESTRING
            yield key + BIN_NONE + value
        elif isinstance(value, float):
            yield BIN_FLOAT32 + key + BIN_NONE + float32.pack(value)
        elif isinstance(value, (COLOR, POINTER, int, int_type)):
            if isinstance(value, COLOR):
                yield BIN_COLOR
            elif isinstance(value, POINTER):
                yield BIN_POINTER
            else:
                yield BIN_INT32
            yield key + BIN_NONE
            yield int32.pack(value)
        else:
            raise TypeError("Unsupported type: %s" % type(value))

    yield BIN_END if not alt_format else BIN_END_ALT


def vbkv_loads(s, mapper=dict, merge_duplicate_keys=True):
    """
    Deserialize ``s`` (``bytes`` containing a VBKV to a Python object.

    ``mapper`` specifies the Python object used after deserializetion. ``dict` is
    used by default. Alternatively, ``collections.OrderedDict`` can be used if you
    wish to preserve key order. Or any object that acts like a ``dict``.

    ``merge_duplicate_keys`` when ``True`` will merge multiple KeyValue lists with the
    same key into one instead of overwriting. You can se this to ``False`` if you are
    using ``VDFDict`` and need to preserve the duplicates.
    """
    if s[:4] != b'VBKV':
        raise ValueError("Invalid header")

    checksum, = struct.unpack('<i', s[4:8])

    if checksum != crc32(s[8:]):
        raise ValueError("Invalid checksum")

    return binary_loads(s[8:], mapper, merge_duplicate_keys, alt_format=True)

def vbkv_dumps(obj):
    """
    Serialize ``obj`` to a VBKV formatted ``bytes``.
    """
    data =  b''.join(_binary_dump_gen(obj, alt_format=True))
    checksum = crc32(data)

    return b'VBKV' + struct.pack('<i', checksum) + data


import logging
import select
import time
from threading import Event as TEvent
from typing import Sequence

from hhd.controller import DEBUG_MODE, Axis, Event, Multiplexer, can_read
from hhd.controller.lib.hide import unhide_all
from hhd.controller.physical.evdev import DINPUT_AXIS_POSTPROCESS, AbsAxis
from hhd.controller.physical.evdev import B as EC
from hhd.controller.physical.evdev import (
    GamepadButton,
    GenericGamepadEvdev,
    enumerate_evs,
    to_map,
)
from hhd.controller.physical.hidraw import GenericGamepadHidraw, enumerate_unique
from hhd.controller.physical.imu import CombinedImu, HrtimerTrigger
from hhd.plugins import Config, Context, Emitter, get_limits, get_outputs

from .const import config_rgb
from .hid import RgbCallback, switch_mode

SELECT_TIMEOUT = 1

logger = logging.getLogger(__name__)

ASUS_VID = 0x0B05
ALLY_PID = 0x1ABE
ALLY_X_PID = 0x1B4C
GAMEPAD_VID = 0x045E
GAMEPAD_PID = 0x028E

ALLY_MAPPINGS: dict[str, tuple[Axis, str | None, float, float | None]] = {
    "accel_x": ("accel_x", "accel", 1, None),
    "accel_y": ("accel_z", "accel", 1, None),
    "accel_z": ("accel_y", "accel", -1, None),
    "anglvel_x": ("gyro_x", "anglvel", 1, None),
    "anglvel_y": ("gyro_z", "anglvel", 1, None),
    "anglvel_z": ("gyro_y", "anglvel", -1, None),
    "timestamp": ("imu_ts", None, 1, None),
}

LIMIT_DEFAULTS = lambda allyx: {
    "s_min": 0 if allyx else 5,
    "s_max": 0x60 if allyx else 0x40,
    "t_min": 5,
    "t_max": 0x60 if allyx else 0x40,
    # ally x vibration motor is too strong
    "vibration": 50 if allyx else 100,
}

MODE_DELAY = 0.15
VIBRATION_DELAY = 0.1
VIBRATION_ON: Event = {
    "type": "rumble",
    "code": "main",
    "strong_magnitude": 0.5,
    "weak_magnitude": 0.5,
}
VIBRATION_OFF: Event = {
    "type": "rumble",
    "code": "main",
    "strong_magnitude": 0,
    "weak_magnitude": 0,
}

FIND_DELAY = 0.1
ERROR_DELAY = 0.3
LONGER_ERROR_DELAY = 3
LONGER_ERROR_MARGIN = 1.3

# TODO: Work with upstream on the xpad (?) driver
# LB = BTN_WEST
# RB = BTN_Z
# X = BTN_C
# A = BTN_SOUTH
# B = BTN_EAST
# Y = BTN_NORTH
# Start (Menu) = BTN_TR
# Select (View) = BTN_TL
# RT = ABS_RZ
# LT = ABS_Z
# L3 = BTN_TL2
# R3 = BTN_TR2
ALLY_X_BUTTON_MAP: dict[int, GamepadButton] = to_map(
    {
        # Gamepad
        "a": [EC("BTN_SOUTH")],
        "b": [EC("BTN_EAST")],
        "x": [EC("BTN_C")],
        "y": [EC("BTN_NORTH")],
        # Sticks
        "ls": [EC("BTN_TL2")],
        "rs": [EC("BTN_TR2")],
        # Bumpers
        "lb": [EC("BTN_WEST")],
        "rb": [EC("BTN_Z")],
        # Select
        "start": [EC("BTN_TR")],
        "select": [EC("BTN_TL")],
        # Misc
        # "mode": [EC("BTN_MODE")],
    }
)

ALLY_X_AXIS_MAP: dict[int, AbsAxis] = to_map(
    {
        # Sticks
        # Values should range from -1 to 1
        "ls_x": [EC("ABS_X")],
        "ls_y": [EC("ABS_Y")],
        "rs_x": [EC("ABS_RX")],
        "rs_y": [EC("ABS_RY")],
        # Triggers
        # Values should range from -1 to 1
        "rt": [EC("ABS_Z")],
        "lt": [EC("ABS_RZ")],
        # Hat, implemented as axis. Either -1, 0, or 1
        "hat_x": [EC("ABS_HAT0X")],
        "hat_y": [EC("ABS_HAT0Y")],
    }
)


class AllyHidraw(GenericGamepadHidraw):
    def __init__(self, *args, kconf={}, rgb_boot, rgb_charging, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.kconf = kconf
        self.mouse_mode = False
        self.rgb_boot = rgb_boot
        self.rgb_charging = rgb_charging
        self.late_init = None

    def open(self) -> Sequence[int]:
        self.queue: list[tuple[Event, float]] = []
        a = super().open()
        if self.dev:
            logger.info(f"Switching Ally Controller to gamepad mode.")
            # Setup leds so they dont interfere after this
            self.dev.write(config_rgb(self.rgb_boot, self.rgb_charging))
            switch_mode(self.dev, "default", self.kconf, first=True)

        self.mouse_mode = False
        self.late_init = time.perf_counter()
        return a

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        # If we can not read return
        if not self.fd or not self.dev:
            return []

        # Process events
        curr = time.perf_counter()
        out: Sequence[Event] = []

        # Remove up to one queued event
        if len(self.queue):
            ev, ofs = self.queue[0]
            if ofs < curr:
                out.append(ev)
                self.queue.pop(0)

        # Force a re-init after 5 seconds in case the MCU did not get the message
        # Hopefully this fixes the back buttons on the og ally.
        if self.late_init and curr > self.late_init + 5:
            logger.info(f"Re-initializing controller.")
            self.late_init = None
            if not self.mouse_mode:
                switch_mode(self.dev, "default", self.kconf, first=True)

        # Read new events
        while can_read(self.fd):
            rep = self.dev.read(self.report_size)
            # logger.warning(f"Received the following report (debug):\n{rep.hex()}")
            if rep[0] != 0x5A:
                continue

            match rep[1]:
                case 0xA6:
                    # action = "left"
                    out.append({"type": "button", "code": "mode", "value": True})
                    self.queue.append(
                        (
                            {"type": "button", "code": "mode", "value": False},
                            curr + MODE_DELAY,
                        )
                    )
                case 0x38:
                    # action = "right"
                    out.append({"type": "button", "code": "share", "value": True})
                    self.queue.append(
                        (
                            {"type": "button", "code": "share", "value": False},
                            curr + MODE_DELAY,
                        )
                    )
                case 0xA7:
                    # right hold
                    # Mode switch
                    if self.mouse_mode:
                        switch_mode(self.dev, "default", self.kconf)
                        self.mouse_mode = False
                        out.append(VIBRATION_ON)
                        self.queue.append((VIBRATION_OFF, curr + VIBRATION_DELAY))
                    else:
                        switch_mode(self.dev, "mouse", self.kconf)
                        self.mouse_mode = True
                        out.append(VIBRATION_ON)
                        self.queue.append((VIBRATION_OFF, curr + VIBRATION_DELAY))
                        self.queue.append((VIBRATION_ON, curr + 2 * VIBRATION_DELAY))
                        self.queue.append((VIBRATION_OFF, curr + 3 * VIBRATION_DELAY))
                case 0xA8:
                    # action = "right_hold_release"
                    pass  # kind of useless

        return out


class AllyXHidraw(GenericGamepadHidraw):
    def open(self) -> Sequence[int]:
        super().open()
        # Drop all events
        return []

    def consume(self, events: Sequence[Event]) -> None:
        if not self.dev:
            return

        for ev in events:
            if ev["type"] != "rumble":
                continue

            if ev["code"] != "main":
                logger.warning(
                    f"Received rumble event with unsupported side: {ev['code']}"
                )
                continue

            "0d 0f 00 00 31 31 ff 00 eb"
            cmd = bytes(
                [
                    0x0D,
                    0x0F,
                    0x00,
                    0x00,
                    min(100, int(ev["weak_magnitude"] * 100)),
                    min(100, int(ev["strong_magnitude"] * 100)),
                    0xFF,
                    0x00,
                    0xEB,
                ]
            )
            self.dev.write(cmd)


def plugin_run(
    conf: Config,
    emit: Emitter,
    context: Context,
    should_exit: TEvent,
    updated: TEvent,
    ally_x: bool,
):
    init = time.perf_counter()
    repeated_fail = False
    first = True
    while not should_exit.is_set():
        try:
            gamepad_devs = enumerate_evs(vid=GAMEPAD_VID)
            nkey_devs = enumerate_unique(vid=ASUS_VID)

            if (not gamepad_devs and not ally_x) or not nkey_devs:
                if first:
                    first = False
                    logger.warning(f"Ally controller not found, waiting...")
                time.sleep(FIND_DELAY)
                continue

            logger.info("Launching emulated controller.")
            updated.clear()
            init = time.perf_counter()
            controller_loop(conf.copy(), should_exit, updated, emit, ally_x)
            repeated_fail = False
        except Exception as e:
            first = True
            failed_fast = init + LONGER_ERROR_MARGIN > time.perf_counter()
            sleep_time = (
                LONGER_ERROR_DELAY if repeated_fail and failed_fast else ERROR_DELAY
            )
            repeated_fail = failed_fast
            logger.error(f"Received the following error:\n{type(e)}:")

            try:
                import traceback

                traceback.print_exc()
            except Exception:
                pass

            logger.error(
                f"Assuming controllers disconnected, restarting after {sleep_time}s."
            )
            # Raise exception
            if DEBUG_MODE:
                raise e
            time.sleep(sleep_time)

    # Unhide all devices before exiting
    unhide_all()


def controller_loop(
    conf: Config, should_exit: TEvent, updated: TEvent, emit: Emitter, ally_x: bool
):
    debug = DEBUG_MODE

    # Output
    d_producers, d_outs, d_params = get_outputs(
        conf["controller_mode"],
        None,
        conf["imu"].to(bool),
        emit=emit,
        rgb_modes={
            "disabled": [],
            "solid": ["color"],
            "pulse": ["color", "speedd"],
            "duality": ["dual", "speedd"],
            "rainbow": ["brightnessd"],
            "spiral": ["brightnessd", "speedd", "direction"],
        },
        rgb_zones="quad",
    )
    motion = d_params.get("uses_motion", True)

    # Imu
    d_imu = CombinedImu(conf["imu_hz"].to(int), ALLY_MAPPINGS, gyro_scale="0.000266")
    d_timer = HrtimerTrigger(conf["imu_hz"].to(int), [HrtimerTrigger.IMU_NAMES])

    # Inputs
    if ally_x:
        d_xinput = GenericGamepadEvdev(
            vid=[ASUS_VID],
            pid=[ALLY_X_PID],
            btn_map=ALLY_X_BUTTON_MAP,
            axis_map=ALLY_X_AXIS_MAP,
            # name=["Generic X-Box pad"],
            capabilities={EC("EV_KEY"): [EC("BTN_A")]},
            required=True,
            postprocess=DINPUT_AXIS_POSTPROCESS,
            hide=True,
        )
        d_allyx = AllyXHidraw(
            vid=[ASUS_VID],
            pid=[ALLY_X_PID],
            usage_page=[0x0F],
            usage=[0x21],
            required=True,
        )
    else:
        d_xinput = GenericGamepadEvdev(
            vid=[GAMEPAD_VID],
            pid=[GAMEPAD_PID],
            # name=["Generic X-Box pad"],
            capabilities={EC("EV_KEY"): [EC("BTN_A")]},
            required=True,
            hide=True,
            postprocess={},  # remove calibration as its supported by the GUI
        )
        d_allyx = None

    # Vendor
    kconf = get_limits(conf["limits"], defaults=LIMIT_DEFAULTS(ally_x))
    d_vend = AllyHidraw(
        vid=[ASUS_VID],
        pid=[ALLY_PID, ALLY_X_PID],
        usage_page=[0xFF31],
        usage=[0x0080],
        required=True,
        rgb_boot=conf.get("rgb_boot", False),
        rgb_charging=conf.get("rgb_charging", False),
        callback=RgbCallback(),
        kconf=kconf,
    )

    # Grab shortcut keyboards
    d_kbd_1 = GenericGamepadEvdev(
        vid=[ASUS_VID],
        pid=[ALLY_PID, ALLY_X_PID],
        capabilities={EC("EV_KEY"): [EC("KEY_F23")]},
        required=True,
        grab=False,
        btn_map={EC("KEY_F17"): "extra_l1", EC("KEY_F18"): "extra_r1"},
    )
    d_kbd_grabbed = False

    multiplexer = Multiplexer(
        trigger="analog_to_discrete",
        dpad="analog_to_discrete",
        share_to_qam=True,
        select_reboots=conf["select_reboots"].to(bool),
        nintendo_mode=conf["nintendo_mode"].to(bool),
        emit=emit,
        swap_guide="select_is_guide" if conf["swap_armory"].to(bool) else None,
        qam_no_release=not conf["swap_armory"].to(bool),
        params=d_params,
    )

    REPORT_FREQ_MIN = 25
    REPORT_FREQ_MAX = 400

    if motion:
        REPORT_FREQ_MAX = max(REPORT_FREQ_MAX, conf["imu_hz"].to(float))

    REPORT_DELAY_MAX = 1 / REPORT_FREQ_MIN
    REPORT_DELAY_MIN = 1 / REPORT_FREQ_MAX

    fds = []
    devs = []
    fd_to_dev = {}

    def prepare(m):
        devs.append(m)
        fs = m.open()
        fds.extend(fs)
        for f in fs:
            fd_to_dev[f] = m

    try:
        d_vend.open()
        prepare(d_xinput)
        if d_allyx:
            prepare(d_allyx)
        if motion:
            if d_timer.open():
                prepare(d_imu)
        prepare(d_kbd_1)
        for d in d_producers:
            prepare(d)

        logger.info("Emulated controller launched, have fun!")
        while not should_exit.is_set() and not updated.is_set():
            start = time.perf_counter()
            # Add timeout to call consumers a minimum amount of times per second
            r, _, _ = select.select(fds, [], [], REPORT_DELAY_MAX)
            evs = []
            to_run = set()
            for f in r:
                to_run.add(id(fd_to_dev[f]))

            for d in devs:
                if id(d) in to_run:
                    evs.extend(d.produce(r))
            evs.extend(d_vend.produce(r))

            evs = multiplexer.process(evs)
            if evs:
                if debug:
                    logger.info(evs)

            d_vend.consume(evs)
            d_xinput.consume(evs)
            if d_allyx:
                d_allyx.consume(evs)

            for d in d_outs:
                d.consume(evs)

            if d_vend.mouse_mode and d_kbd_grabbed and d_kbd_1.dev:
                try:
                    d_kbd_1.dev.ungrab()
                except Exception:
                    pass
                d_kbd_grabbed = False
            elif not d_vend.mouse_mode and not d_kbd_grabbed and d_kbd_1.dev:
                try:
                    d_kbd_1.dev.grab()
                except Exception:
                    pass
                d_kbd_grabbed = True

            # If unbounded, the total number of events per second is the sum of all
            # events generated by the producers.
            # For Legion go, that would be 100 + 100 + 500 + 30 = 730
            # Since the controllers of the legion go only update at 500hz, this is
            # wasteful.
            # By setting a target refresh rate for the report and sleeping at the
            # end, we ensure that even if multiple fds become ready close to each other
            # they are combined to the same report, limiting resource use.
            # Ideally, this rate is smaller than the report rate of the hardware controller
            # to ensure there is always a report from that ready during refresh
            t = time.perf_counter()
            elapsed = t - start
            if elapsed < REPORT_DELAY_MIN:
                time.sleep(REPORT_DELAY_MIN - elapsed)

    except KeyboardInterrupt:
        raise
    finally:
        try:
            d_vend.close(not updated.is_set())
        except Exception as e:
            logger.error(f"Error while closing device '{d}' with exception:\n{e}")
            if debug:
                raise e
        try:
            d_timer.close()
        except Exception as e:
            logger.error(f"Error while closing device '{d}' with exception:\n{e}")
            if debug:
                raise e
        for d in reversed(devs):
            try:
                d.close(not updated.is_set())
            except Exception as e:
                logger.error(f"Error while closing device '{d}' with exception:\n{e}")
                if debug:
                    raise e



from hhd.logging import RASTER
import sys
import subprocess

IMPORTANT_FILES = {
    "/sys/class/dmi/id/board_vendor": "Board Vendor",
    "/sys/class/dmi/id/board_name": "Board Name",
    "/sys/class/dmi/id/board_version": "Board Version",
    "/sys/class/dmi/id/product_family": "Product Family",
    "/sys/class/dmi/id/product_name": "Product Name",
    "/sys/class/dmi/id/sys_vendor": "System Vendor",
    "/sys/class/dmi/id/modalias": "DMI Modalias",
    "/sys/class/dmi/id/bios_version": "BIOS Version",
    "/sys/class/dmi/id/ec_firmware_release": "EC Firmware",
}

JOURNAL_FIRST_LINES = 1_000
JOURNAL_MAX_SIZE = 8_000

JOURNALCTL_CMD = lambda boot: [
    "journalctl",
    "--no-pager",
    "--no-hostname",
    "-b",
    f"{boot}",
]

JOURNALCTL_BLACKLIST = [
    "vivaldi",
]


def get_log(boot: int) -> str:
    import datetime

    out = RASTER

    out += f"""
Debug Log created by Handheld Daemon at {datetime.datetime.now().strftime('%d/%m/%Y, %H:%M:%S')}.

"""

    out += "# Device Information\n"

    for path, name in IMPORTANT_FILES.items():
        try:
            with open(path, "r") as f:
                out += f'{name} ({path}):\n"{f.read().strip()}"\n'
        except Exception as e:
            out += f"Error reading {name} ({path}): {e}\n"

    out += "Kernel Version (uname -sr):\n"
    try:
        out += f"\"{subprocess.run(
            ['uname', '-sr'], capture_output=True, text=True
        ).stdout.strip()}\"\n"
    except Exception as e:
        out += f"Error reading kernel version: {e}\n"

    out += "\n# OS Release\n"
    try:
        with open("/etc/os-release", "r") as f:
            out += f"{f.read()}"
    except Exception as e:
        out += f"Error reading /etc/os-release: {e}\n"

    out += f"\n\n# Journalctl from boot index {boot}\n"
    try:
        lines = list(
            subprocess.run(
                JOURNALCTL_CMD(boot), capture_output=True, text=True
            ).stdout.splitlines()
        )

        # Write last lines last to allow truncating the middle
        written = []
        for line in reversed(
            lines[max(JOURNAL_FIRST_LINES, len(lines) - JOURNAL_MAX_SIZE) :]
        ):
            if not any(bl in line for bl in JOURNALCTL_BLACKLIST):
                written.append(line)
                if len(written) >= JOURNAL_MAX_SIZE:
                    written.append("\n... (truncated)\n")
                    break

        # Write first lines
        for line in reversed(lines[:JOURNAL_FIRST_LINES]):
            if not any(bl in line for bl in JOURNALCTL_BLACKLIST):
                written.append(line)

        out += "\n".join(reversed(written))

    except Exception as e:
        out += f"Error reading journalctl: {e}\n"

    return out


if __name__ == "__main__":
    print(get_log(0))



import logging
import os
import subprocess
import time
from select import select
from threading import Event as TEvent
from typing import Any, NamedTuple, Sequence

import Xlib
from Xlib import XK, X, Xatom, display, error
from Xlib.ext.xtest import fake_input

from hhd.plugins import Context, Emitter, Config
from hhd.utils import restore_priviledge, switch_priviledge

logger = logging.getLogger(__name__)

X11_DIR = b"/tmp/.X11-unix/"
HHD_ID = 5335
STEAM_ID = 769


class CachedValues(NamedTuple):
    overlay: bool
    focus: bool
    notify: bool
    touch: int | None


QAM_DELAY = 0.35


class QamHandlerGamescope:
    def __init__(
        self, ctx=None, force_disp: str | None = None, compat_send: bool = True
    ) -> None:
        self.disp = None
        self.ctx = ctx
        self.force_disp = force_disp
        self.compat_send = compat_send

    def _register_display(self):
        self.close()
        try:
            if self.force_disp:
                res = display.Display(self.force_disp), self.force_disp
            else:
                res = get_overlay_display(get_gamescope_displays(), self.ctx)
            if not res:
                logger.info(
                    f"Could not find gamescope display, sending compatibility QAM."
                )
                return False
            self.disp, name = res
            logger.info(f"Registering display {name} to send QAM events to.")
            return True
        except Exception as e:
            logger.info(f"Error while registering Gamescope display for QAM:\n{e}.")
            return False

    def _send_qam(self, expanded=False):
        try:
            disp = self.disp
            if not disp:
                return False
            get_key = lambda k: disp.keysym_to_keycode(XK.string_to_keysym(k))
            KCTRL = get_key("Control_L")
            KEY = get_key("1" if expanded else "2")

            # Checking for steam seemed to work, but blanket sending the command
            # with a compatibility QAM at first works the same and this looks
            # fragile
            # steam = find_steam(disp)
            # if not steam:
            #     logger.info(f"Could not find Steam (?). Sending compatibility QAM.")
            #     return False

            fake_input(disp, X.KeyPress, KCTRL)  # , root=steam)
            fake_input(disp, X.KeyPress, KEY)  # , root=steam)
            disp.sync()
            time.sleep(QAM_DELAY)
            fake_input(disp, X.KeyRelease, KCTRL)  # , root=steam)
            fake_input(disp, X.KeyRelease, KEY)  # , root=steam)
            disp.sync()
            logger.info(f"Sent QAM event directly to gamescope.")
            return True
        except Exception as e:
            logger.warning(
                f"Could not send QAM to Gamescope with error:\n{e}\nSending compatibility QAM."
            )
            return False

    def __call__(self, expanded=False) -> Any:
        if self._send_qam(expanded):
            return True
        # Steam fails to open QAM with ctrl+2 the first time
        # So send compatibility QAM if we have to register display
        if self._register_display() and self.compat_send:
            logger.info(
                "Sending compatibility QAM as first QAM, as display was registered now."
            )
        if not self.compat_send:
            return self._send_qam(expanded)
        return False

    def close(self):
        if self.disp:
            try:
                self.disp.close()
                self.disp = None
            except Exception:
                pass


def find_x11_auth(ctx: Context):
    # TODO: Fix hardcoding runtime dir
    LOCATION = f"/run/user/{ctx.euid}"
    for fn in sorted(os.listdir(LOCATION)):
        if (
            # KDE
            fn.startswith("xauth_")
            # GNOME
            or fn.startswith(".mutter-Xwaylandauth.")
        ):
            return os.path.join(LOCATION, fn)


def find_x11_display(ctx: Context):
    for fn in sorted(os.listdir(X11_DIR)):
        if fn and os.stat(X11_DIR + fn).st_uid == ctx.euid:
            return ":" + fn[1:].decode()


def get_gamescope_displays():
    """Returns X11 UNIX sockets from gamescope opened under /tmp"""
    files = subprocess.run(["lsof", "-c", "gamescope-wl", "-Fn"], capture_output=True)
    out = []
    for ln in files.stdout.splitlines():
        if len(ln) < 1:
            continue
        ln = ln[1:]

        if not ln.startswith(X11_DIR):
            continue

        fn = ln.split(b" ")[0]
        disp = ":" + fn[len(X11_DIR) + 1 :].decode()
        out.append(disp)
    return out


def is_gamescope_running():
    return bool(get_gamescope_displays())


def get_overlay_display(displays: Sequence[str], ctx=None):
    """Probes the provided gamescope displays to find the overlay one."""

    # FIXME: Fix authentication without priviledge deescalation
    if ctx:
        old = switch_priviledge(ctx, False)
    else:
        old = None

    try:
        for disp in displays:
            try:
                d = display.Display(disp)

                atoms = [d.get_atom_name(v) for v in d.screen().root.list_properties()]
                if "GAMESCOPE_FOCUSED_WINDOW" in atoms:
                    return d, disp

                d.close()
            except Exception:
                pass
    finally:
        if old:
            restore_priviledge(old)


def apply_gamescope_config(display: display.Display, config: Config, prev: dict):
    apply = False
    
    halfhz = config.get("steamui_halfhz", None)
    halfhz_rev = prev.get("steamui_halfhz", None)
    if halfhz is not None and halfhz != halfhz_rev:
        display.screen().root.change_property(
            display.get_atom("GAMESCOPE_STEAMUI_HALFHZ"), Xatom.CARDINAL, 32, [int(halfhz)]
        )
        logger.info(f"Setting SteamUI halfhz to {halfhz}.")
        prev["steamui_halfhz"] = halfhz
        apply = True
    
    if apply:
        display.flush()

def find_wins(display: display.Display, win: list[str], atoms: list[str] = []):
    n = display.get_atom("WM_CLASS")
    a_ids = [display.get_atom(a, only_if_exists=True) for a in atoms]

    wins = []
    for w in display.screen().root.query_tree().children:
        # Check the window has the proper class
        v = w.get_property(n, Xatom.STRING, 0, 50)
        if not v:
            continue
        if not v.value:
            continue

        # Check the window has all the required atoms
        for a_id in a_ids:
            if not w.get_property(a_id, Xatom.STRING, 0, 50):
                return

        classes = [c.decode() for c in v.value.split(b"\00") if c]
        found = True
        for val in win:
            if val not in classes:
                found = False

        if found:
            wins.append(w)
    return wins


def find_win(display: display.Display, win: list[str], atoms: list[str] = []):
    out = find_wins(display, win, atoms)
    return out[0] if out else None


def register_changes(display, win):
    win.change_attributes(event_mask=Xlib.X.PropertyChangeMask)
    display.flush()
    display.sync()


def find_hhd(display: display.Display):
    return find_win(display, ["dev.hhd.hhd-ui"])


def find_steam(display: display.Display):
    return find_win(display, ["steamwebhelper", "steam"]) or find_win(
        display, ["steamwebhelper", "SDL Application"]
    )


def does_steam_exist(display: display.Display):
    return find_win(display, ["steamwebhelper"]) or find_win(display, ["steam"])


def print_data(display: display.Display):
    for w in (find_hhd(display), find_steam(display), display.screen().root):
        if not w:
            continue
        for p in w.list_properties():
            req = w.get_property(p, Xatom.CARDINAL, 0, 100)
            if req:
                v = list(req.value) if req.value else None
            else:
                v = None
            print(f"{p:4d}-{display.get_atom_name(p):>40s}: {v}")
        print()


def print_debug(display: display.Display, args: list[str] = []):
    d = display
    r = display.screen().root

    if "noatoms" not in args:
        print("ATOMS:")
        for v in r.list_properties():
            print(f"{v: 4d}: {d.get_atom_name(v)}")

    if "root" in args:
        windows = [r]
    else:
        windows = [r, *r.query_tree().children]

    print()
    print("WINDOWS:")
    for i, w in enumerate(windows):
        print(f"\n{i:02d}:", end="")
        for p in w.list_properties():
            n = d.get_atom_name(p)
            if "WM_NAME" == n:
                print(f" '{w.get_property(p, Xatom.STRING, 0, 100).value.decode()}'")
                break
        else:
            print(" no name")

        for p in w.list_properties():
            n = d.get_atom_name(p)
            if "STEAM" in n or "GAMESCOPE" in n:
                print(
                    f"> {n}: {list(w.get_property(p, Xatom.CARDINAL, 0, 15).value)},",
                )
        for p in w.list_properties():
            n = d.get_atom_name(p)
            if "STEAM" not in n and "GAMESCOPE" not in n:
                print(
                    f"- {n}: {list(w.get_property(p, Xatom.CARDINAL, 0, 15).value) or w.get_property(p, Xatom.STRING, 0, 15).value},",
                )


def prepare_hhd(display, hhd, steam=None):
    if not steam:
        # If hhd appears a game steam will have issues with per-game profiles
        hhd.change_property(
            display.get_atom("STEAM_GAME"), Xatom.CARDINAL, 32, [HHD_ID]
        )
    hhd.change_property(display.get_atom("STEAM_NOTIFICATION"), Xatom.CARDINAL, 32, [0])
    hhd.change_property(display.get_atom("STEAM_BIGPICTURE"), Xatom.CARDINAL, 32, [1])
    hhd.change_property(display.get_atom("GAMESCOPE_NO_FOCUS"), Xatom.CARDINAL, 32, [1])
    display.flush()
    display.sync()


def process_events(disp):
    try:
        found = False
        for _ in range(disp.pending_events()):
            ev = disp.next_event()
            if ev and hasattr(ev, "atom") and "STEAM" in disp.get_atom_name(ev.atom):
                found = True
        return found
    except Exception as e:
        logger.warning(f"Failed to process display events with error:\n{e}")
    return True

def get_current_game(display):
    stat_game = display.get_atom("GAMESCOPE_FOCUSED_APP_GFX")
    game = display.screen().root.get_property(stat_game, Xatom.CARDINAL, 0, 15)
    return game.value[0] if game and game.value else None


def set_dpms(display, enable: bool):
    stat_dpms = display.get_atom("GAMESCOPE_DPMS")
    display.screen().root.change_property(
        stat_dpms, Xatom.CARDINAL, 32, [1 if enable else 0]
    )
    display.flush()


def update_steam_values(display, steam, old: CachedValues | None):
    stat_focus = display.get_atom("STEAM_INPUT_FOCUS")
    stat_overlay = display.get_atom("STEAM_OVERLAY")
    stat_notify = display.get_atom("STEAM_NOTIFICATION")
    stat_click = display.get_atom("STEAM_TOUCH_CLICK_MODE")

    def was_set(v):
        prop = steam.get_property(v, Xatom.CARDINAL, 0, 15)
        return prop and prop.value and bool(prop.value[0])

    new_focus = was_set(stat_focus)
    new_overlay = was_set(stat_overlay)
    new_notify = was_set(stat_notify)

    # Use some weird logic to get previous touch value
    # Essentially, only remember that value if it was set and different than TARGET_TOUCH
    r = display.screen().root
    prop = r.get_property(stat_click, Xatom.CARDINAL, 0, 15)
    touch_was_set = prop and prop.value and prop.value[0] != TARGET_TOUCH
    touch_val = prop.value[0] if touch_was_set else None
    if touch_val is None and old and old.touch is not None:
        touch_val = old.touch

    out = CachedValues(
        focus=new_focus or (old.focus if old else False),
        overlay=new_overlay or (old.overlay if old else False),
        notify=new_notify or (old.notify if old else False),
        touch=touch_val,
    )
    return out, new_focus or new_overlay or new_notify


TARGET_TOUCH = 4


def show_hhd(display, hhd, steam):
    stat_focus = display.get_atom("STEAM_INPUT_FOCUS")
    stat_overlay = display.get_atom("STEAM_OVERLAY")
    stat_notify = display.get_atom("STEAM_NOTIFICATION")
    stat_click = display.get_atom("STEAM_TOUCH_CLICK_MODE")

    # Unfortunately, doing the commented out section breaks steam profiles
    # and enables desktop mode steam input on Handheld Daemon, showing a mouse

    # # Here, we do a bit of trickery with steam
    # # We pretend to be one of the games that the user has launched to not break
    # # steam profiles and to get steam to ignore its input
    # stat_game = display.get_atom("STEAM_GAME")
    # stat_focusable = display.get_atom("GAMESCOPE_FOCUSABLE_APPS")

    # If steam set the touch value to something else, try to override it with 1
    r = display.screen().root
    prop = r.get_property(stat_click, Xatom.CARDINAL, 0, 15)
    touch_was_set = prop and prop.value

    hhd.change_property(stat_focus, Xatom.CARDINAL, 32, [1])
    hhd.change_property(stat_overlay, Xatom.CARDINAL, 32, [1])
    if steam:
        steam.change_property(stat_focus, Xatom.CARDINAL, 32, [0])
        steam.change_property(stat_overlay, Xatom.CARDINAL, 32, [0])
        steam.change_property(stat_notify, Xatom.CARDINAL, 32, [0])

        # # Use a game id for hhd so that steam does not leak input
        # new_id = HHD_ID
        # focusable = display.screen().root.get_property(
        #     stat_focusable, Xatom.CARDINAL, 0, 50
        # )
        # if focusable and focusable.value:
        #     for i in focusable.value:
        #         if i == HHD_ID and i != STEAM_ID:
        #             new_id = i
        #             break
        # logger.info(f"Setting HHD as game '{new_id}' to disable steam navigation.")
        # hhd.change_property(stat_game, Xatom.CARDINAL, 32, [new_id])

    if touch_was_set:
        # Give it a bit of time before setting the touch target to avoid steam
        # messing with it
        display.flush()
        display.sync()
        time.sleep(0.1)
        r.change_property(stat_click, Xatom.CARDINAL, 32, [TARGET_TOUCH])

    display.flush()
    display.sync()


def hide_hhd(display, hhd, steam, old: CachedValues | None):
    stat_focus = display.get_atom("STEAM_INPUT_FOCUS")
    stat_overlay = display.get_atom("STEAM_OVERLAY")
    stat_notify = display.get_atom("STEAM_NOTIFICATION")
    stat_click = display.get_atom("STEAM_TOUCH_CLICK_MODE")

    # Set values
    hhd.change_property(stat_focus, Xatom.CARDINAL, 32, [0])
    hhd.change_property(stat_overlay, Xatom.CARDINAL, 32, [0])

    # Restore steam
    if steam and old:
        if old.focus:
            steam.change_property(stat_focus, Xatom.CARDINAL, 32, [1])
        if old.overlay:
            steam.change_property(stat_overlay, Xatom.CARDINAL, 32, [1])
        if old.notify:
            steam.change_property(stat_notify, Xatom.CARDINAL, 32, [1])
        if old.touch is not None:
            display.screen().root.change_property(
                stat_click, Xatom.CARDINAL, 32, [old.touch]
            )

    display.flush()
    display.sync()


def find_focusable_windows(display):
    stat_focusable = display.get_atom("GAMESCOPE_FOCUSABLE_APPS")
    focusable = display.screen().root.get_property(
        stat_focusable, Xatom.CARDINAL, 0, 50
    )
    return focusable.value if focusable and focusable.value else []


def make_hhd_not_focusable(display):
    stat_focused = display.get_atom("GAMESCOPECTRL_BASELAYER_APPID")
    stat_focusable = display.get_atom("GAMESCOPE_FOCUSABLE_APPS")

    focusable = display.screen().root.get_property(
        stat_focusable, Xatom.CARDINAL, 0, 50
    )
    curr = display.screen().root.get_property(stat_focused, Xatom.CARDINAL, 0, 50)

    if not focusable or not focusable.value:
        # Cannot print here or the logs will be swarmed
        # There should always be something here
        return

    # Check whether we should write focusable apps to hide hhd
    write_focus = False
    if not curr or not curr.value:
        write_focus = True
    else:
        for i in focusable.value:
            if i == HHD_ID:
                # skip hhd
                continue
            found = False
            for j in curr.value:
                if i == j:
                    found = True
                    break
            if not found:
                write_focus = True
                break

    # Hide HHD
    if write_focus:
        new_focus = [v for v in focusable.value if v != HHD_ID]
        logger.info(
            f"Hiding Handheld Daemon from gamescope. Setting focusable apps to: {new_focus}"
        )
        display.screen().root.change_property(
            stat_focused, Xatom.CARDINAL, 32, new_focus
        )
        display.flush()
        display.sync()


def monitor_gamescope(emit: Emitter, ctx, should_exit: TEvent):
    GAMESCOPE_WAIT = 2
    GAMESCOPE_GUARD = 1

    should_exit = TEvent()

    while not should_exit.is_set():
        # Wait for gamescope
        try:
            res = get_overlay_display(get_gamescope_displays(), ctx)
            if not res:
                time.sleep(GAMESCOPE_WAIT)
                continue

            d, name = res
            logger.info(f"Found gamescope display {name}")
            r = d.screen().root
            r.change_attributes(event_mask=X.PropertyChangeMask)
            fn = d.fileno()
            atom = d.get_atom("GAMESCOPE_FOCUSED_APP_GFX")
            old = None

            while not should_exit.is_set():
                rs = select([fn], [], [], GAMESCOPE_GUARD)[0]
                if not rs:
                    continue

                process_events(d)

                val = r.get_property(atom, Xatom.CARDINAL, 0, 5)
                if not val or not val.value:
                    continue

                game = val.value[0]
                if old != game:
                    old = game
                    logger.warning(game)

        except Exception as e:
            logger.warning(f"Lost connection to gamescope. Did steam exit? Error:\n{e}")
            time.sleep(GAMESCOPE_WAIT)



# import os

# def get_game_data(appcache: str):from hhd.plugins.overlay.steam import appcache
import os

from .appcache import parse_appinfo

def get_games(appdir: str):
    with open(os.path.join(appdir, "appinfo.vdf"), 'rb') as f:
        games = {}
        _, data = parse_appinfo(f)
        for d in data:
            try:
                appid = str(d['appid'])
                name = d['data']['appinfo']['common']['name']
                games[appid] = {"name": name, "images": []}
            except KeyError:
                pass
    
    images = {}
    libdir = os.path.join(appdir, "librarycache")
    for fn in os.listdir(libdir):
        try:
            id_split = fn.index('_')
            ext_split = fn.rindex('.')
            appid = fn[:id_split]
            itype = fn[id_split+1:ext_split]

            if appid not in games:
                continue

            games[appid]["images"].append(itype)

            if appid not in images:
                images[appid] = {}

            images[appid][itype] = os.path.join(libdir, fn)
        except ValueError:
            pass
    
    return games, images


import argparse


def main():
    parser = argparse.ArgumentParser(
        prog="HHD: Handheld Daemon contribution helper scripts",
        description="Scripts to automate the capture of events, etc.",
    )
    parser.add_argument(
        "command",
        nargs="+",
        default=[],
        help="Supported commands: `evdev`, `hidraw`, `gamescope`",
    )
    args = parser.parse_args()

    cmds = args.command
    try:
        match cmds[0]:
            case "evdev":
                from .dev import evdev

                evdev(cmds[1] if len(cmds) > 1 else None)
            case "hidraw":
                from .dev import hidraw

                if len(cmds) > 1:
                    hidraw(*cmds[1:])
                else:
                    hidraw(None)
            case "gamescope":
                from .gs import gamescope_debug

                gamescope_debug(cmds[1:])
            case _:
                print(f"Command `{cmds[0]}` not supported.")
    except KeyboardInterrupt:
        pass



import logging
import os
from threading import Event as TEvent
from threading import Thread
from typing import Sequence

from hhd.plugins import Config, Context, Event, HHDPlugin, load_relative_yaml
from hhd.utils import expanduser

from ..plugin import open_steam_kbd
from .const import get_system_info, get_touchscreen_quirk
from .controllers import QamHandlerKeyboard, device_shortcut_loop
from .steam import get_games
from .x11 import is_gamescope_running

logger = logging.getLogger(__name__)

SHORTCUT_RELOAD_DELAY = 2

HHD_OVERLAY_DISABLE = os.environ.get("HHD_OVERLAY_DISABLE", "0") == "1"
FORCE_GAME = os.environ.get("HHD_FORCE_GAME_ID", None)
SUPPORTS_HALVING = os.environ.get("HHD_GS_STEAMUI_HALFHZ", "0") == "1"
SUPPORTS_DPMS = os.environ.get("HHD_GS_DPMS", "0") == "1"


def load_steam_games(ctx: Context, emit, burnt_ids: set):
    # Defer loading until we enter a game
    info = emit.info
    curr = info.get("game.id", None)
    # Lump steam into none to avoid loading twice
    if info.get("game.is_steam", False):
        curr = None

    # If the game changes and we do not have data for it do a reload
    if curr in burnt_ids:
        return None, None

    if "games" in info and curr in info.get("games", {}):
        return None, None

    # Maybe a game is missing from appcache, if it is burn it
    # so we dont try to load the library again
    burnt_ids.add(curr)

    try:
        # Load the games
        games, images = get_games(expanduser("~/.local/share/Steam/appcache/", ctx))
        logger.info(f"Loaded info for {len(games)} steam games.")

        # Add correct game data after refreshing the database (e.g., the user
        # downloaded a new game)
        if curr and curr in games:
            emit.info["game.data"] = games[curr]

        return games, images
    except Exception as e:
        logger.warning(f"Could not load steam games:\n{e}")
        return None, None


class OverlayPlugin(HHDPlugin):
    def __init__(self) -> None:
        self.name = f"overlay"
        self.priority = 75
        self.log = "ovrl"
        self.ovf = None
        self.initialized = False
        self.old_shortcuts = None
        self.short_should_exit = None
        self.has_correction = True
        self.old_touch = False
        self.old_asus_cycle = None
        self.short_t = None
        self.has_executable = False
        self.qam_handler = None
        self.qam_handler_fallback = None
        self.touch_gestures = True
        self.ctx = None

        self.images = None
        self.burnt_ids = set()

    def open(
        self,
        emit,
        context: Context,
    ):
        try:
            from .base import OverlayService
            from .overlay import find_overlay_exe
            from .x11 import QamHandlerGamescope

            self.ovf = OverlayService(context, emit)
            self.ctx = context
            self.has_executable = bool(find_overlay_exe(context))

            if bool(os.environ.get("HHD_QAM_KEYBOARD", None)):
                # Sends the events as ctrl+1, ctrl+2
                self.qam_handler = QamHandlerKeyboard()
            elif bool(os.environ.get("HHD_QAM_GAMESCOPE", None)):
                # Sends X11 events to gamescope. Stopped working after libei
                self.qam_handler = QamHandlerGamescope(context)
            else:
                self.qam_handler = None

            if self.qam_handler:
                emit.register_qam(self.qam_handler)
            else:
                self.qam_handler_fallback = QamHandlerKeyboard()
            self.emit = emit
        except Exception as e:
            logger.warning(
                f"Could not init overlay service, is python-xlib installed? Error:\n{e}"
            )
            self.ovf = None

    def settings(self):
        if not self.ovf:
            return {}

        self.initialized = True
        set = {
            "gamemode": load_relative_yaml("gamemode.yml"),
            "shortcuts": load_relative_yaml("shortcuts.yml"),
        }

        if not SUPPORTS_HALVING:
            del set["gamemode"]["gamescope"]["children"]["steamui_halfhz"]
        if not SUPPORTS_DPMS:
            del set["gamemode"]["gamescope"]["children"]["dpms"]

        if get_touchscreen_quirk(None, None)[0] and not os.environ.get(
            "HHD_ALLOW_CORRECTION", None
        ):
            # For devices with a dmi match, hide orientation correction
            self.has_correction = False
            del set["shortcuts"]["touchscreen"]["children"]["orientation"]
        else:
            self.has_correction = True
            set["shortcuts"]["touchscreen"]["children"]["orientation"]["modes"][
                "manual"
            ]["children"]["dmi"]["default"] = " - ".join(
                map(lambda x: f'"{x}"', get_system_info())
            )
        return set

    def update(self, conf: Config):
        self.emit.set_simple_qam(not self.has_executable)

        # Load game information
        if self.ctx:
            games, images = load_steam_games(self.ctx, self.emit, self.burnt_ids)
            if games and images:
                self.emit.set_gamedata(games, images)
        if FORCE_GAME:
            self.emit.info["game.id"] = FORCE_GAME
            self.emit.info["game.is_steam"] = False
            self.emit.info["game.data"] = self.emit.get_gamedata(FORCE_GAME)
        if self.ovf:
            self.ovf.launch_overlay()

        self.touch_gestures = not bool(
            conf.get("gamemode.display.gestures_disable", False)
        )
        if SUPPORTS_HALVING and self.ovf:
            self.ovf.gsconf["steamui_halfhz"] = conf.get(
                "gamemode.gamescope.steamui_halfhz", False
            )
        if SUPPORTS_DPMS and self.ovf:
            self.ovf.gsconf["dpms"] = conf.get("gamemode.gamescope.dpms", False)

        disable_touch = conf.get("gamemode.display.touchscreen_disable", False)
        if disable_touch is None:
            # Initialize value since there is no default
            disable_touch = False
            conf["gamemode.display.touchscreen_disable"] = False

        asus_cycle = conf.get("tdp.asus.cycle_tdp", False)
        if self.initialized and (
            not self.old_shortcuts
            or self.old_shortcuts != conf["shortcuts"]
            or self.old_touch != disable_touch
            or self.old_asus_cycle != asus_cycle
        ):
            self.old_asus_cycle = asus_cycle
            self.old_shortcuts = conf["shortcuts"].copy()
            self._close_short()

            kbd = False
            for v in ("meta_press", "meta_hold", "ctrl_3", "ctrl_4"):
                kbd = (
                    kbd or conf.get(f"shortcuts.keyboard.{v}", "disabled") != "disabled"
                )
            touch = False
            for v in ("bottom", "left_top", "left_bottom", "right_top", "right_bottom"):
                touch = (
                    touch
                    or conf.get(f"shortcuts.touchscreen.{v}", "disabled") != "disabled"
                )
            # ctrl = (
            #     conf.get("shortcuts.controller.xbox_b", "disabled") != "disabled"
            #     or asus_cycle
            # )
            # For now always monitor controllers to be able to grab
            ctrl = True
            # if self.ovf:
            #     self.ovf.interceptionSupported = True

            if kbd or touch or ctrl or disable_touch:
                logger.info(
                    f"Starting shortcut loop with:\nkbd: {kbd}, touch: {touch}, ctrl: {ctrl}, disable_touch: {disable_touch}"
                )
                self.short_should_exit = TEvent()
                touch_correction = (
                    conf.get("shortcuts.touchscreen.orientation.manual", None)
                    if self.has_correction
                    and conf.get("shortcuts.touchscreen.orientation.mode", "auto")
                    == "manual"
                    else None
                )
                self.short_t = Thread(
                    target=device_shortcut_loop,
                    args=(
                        self.emit,
                        self.short_should_exit,
                        False,
                        kbd,
                        ctrl,
                        touch,
                        disable_touch,
                        touch_correction,
                    ),
                )
                self.short_t.start()
                self.old_touch = disable_touch
            else:
                logger.info("No shortcuts enabled, not starting shortcut loop.")

    def notify(self, events: Sequence[Event]):
        if self.ovf:
            self.ovf.notify(events)

        for ev in events:
            if ev["type"] != "special":
                continue

            side = None
            section = None
            override_enable = False
            match ev["event"]:
                case gesture if gesture.startswith("swipe_"):
                    if self.touch_gestures:
                        side = gesture[len("swipe_") :]
                        section = "touchscreen"
                    cmd = None
                case gesture if gesture.startswith("kbd_"):
                    if is_gamescope_running():
                        # Only allow kbd shortcuts while gamescope is open
                        # Cannot be used in big picture because KDE/GNOME
                        side = gesture[len("kbd_") :]
                        section = "keyboard"
                    cmd = None
                case "xbox_b":
                    side = "xbox_b"
                    section = "controller"
                case "xbox_y":
                    side = "xbox_y"
                    section = "controller"
                case "qam_hold":
                    # Open QAM with hold for accessibility
                    cmd = "open_qam"
                case "qam_predouble":
                    cmd = "open_qam_if_closed"
                case "qam_double":
                    # Preferred bind for QAM is dual press
                    cmd = "open_qam"
                case "overlay":
                    override_enable = True
                    cmd = "open_qam"
                case "qam_triple":
                    # Allow opening expanded menu with tripple press
                    cmd = "open_expanded"
                case _:
                    cmd = None

            if section and side and self.old_shortcuts:
                logger.info(f"Gesture: {ev['event']}, section: {section}, key: {side}")
                cmd_raw = self.old_shortcuts.get(f"{section}.{side}", "disabled")
                cmd = None
                match cmd_raw:
                    case "disconnect":
                        d = ev.get("data", None)
                        uniq = d.get("uniq", None) if d else None
                        import re

                        # Make sure uniq is kind of a mac address
                        # We are a root level daemon
                        if uniq and re.match(r"([\d:]+)", uniq):
                            logger.warning(
                                f"Disconnecting controller with uniq: {uniq}"
                            )
                            os.system("bluetoothctl disconnect " + uniq)
                    case "hhd_qam":
                        cmd = "open_qam"
                    case "hhd_expanded":
                        cmd = "open_expanded"
                    case "steam_qam":
                        logger.info("Opening steam qam.")
                        if (
                            not self.emit.open_steam(False)
                            and self.qam_handler_fallback
                        ):
                            self.qam_handler_fallback(False)
                    case "steam_expanded":
                        logger.info("Opening steam expanded.")
                        if not self.emit.open_steam(True) and self.qam_handler_fallback:
                            self.qam_handler_fallback(True)
                    case "keyboard":
                        if open_steam_kbd(self.emit, True):
                            logger.info("Opened Steam keyboard.")
                        else:
                            logger.warning(
                                "Could not open Steam keyboard. Is Steam running?"
                            )
                    case "screenshot":
                        logger.info("Taking screenshot.")
                        if self.qam_handler and hasattr(self.qam_handler, "screenshot"):
                            getattr(self.qam_handler, "screenshot")()
                        elif self.qam_handler_fallback:
                            self.qam_handler_fallback.screenshot()

            if self.ovf and cmd:
                init = "close" not in cmd
                if init:
                    logger.info(f"Executing overlay command: '{cmd}'")
                self.ovf.update(cmd, init)

    def _close_short(self):
        if self.short_should_exit:
            self.short_should_exit.set()
            self.short_should_exit = None
        if self.short_t:
            self.short_t.join()
            self.short_t = None

    def close(self):
        if self.ovf:
            self.ovf.close()
        if self.qam_handler:
            self.qam_handler.close()
        if self.qam_handler_fallback:
            self.qam_handler_fallback.close()
        self._close_short()


def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if HHD_OVERLAY_DISABLE:
        return []

    if len(existing):
        return existing

    return [OverlayPlugin()]
